from sickchill import settings
from sickchill.oldbeard import common


class Notifier(object):
    def notify_snatch(self, ep_name):
        if settings.DISCORD_NOTIFY_SNATCH:
            self._notify_discord(common.notifyStrings[common.NOTIFY_SNATCH] + ": " + ep_name)

    def notify_download(self, ep_name):
        if settings.DISCORD_NOTIFY_DOWNLOAD:
            self._notify_discord(common.notifyStrings[common.NOTIFY_DOWNLOAD] + ": " + ep_name)

    def notify_subtitle_download(self, ep_name, lang):
        if settings.DISCORD_NOTIFY_SUBTITLEDOWNLOAD:
            self._notify_discord(common.notifyStrings[common.NOTIFY_SUBTITLE_DOWNLOAD] + " " + ep_name + ": " + lang)

    def notify_git_update(self, new_version="??"):
        if settings.USE_DISCORD:
            update_text = common.notifyStrings[common.NOTIFY_GIT_UPDATE_TEXT]
            title = common.notifyStrings[common.NOTIFY_GIT_UPDATE]
            self._notify_discord(title + " - " + update_text + new_version)

    def notify_login(self, ipaddress=""):
        if settings.USE_DISCORD:
            update_text = common.notifyStrings[common.NOTIFY_LOGIN_TEXT]
            title = common.notifyStrings[common.NOTIFY_LOGIN]
            self._notify_discord(title + " - " + update_text.format(ipaddress))

    def test_notify(self, webhook: str = None, name: str = None, avatar: str = None, tts=None):
        from sickchill.oldbeard.notifications_queue import DiscordTask

        task = DiscordTask("This is a test notification from SickChill")
        return task._send_discord(webhook=webhook, name=name, avatar=avatar, tts=tts)

    @staticmethod
    def _send_discord(message=None, force=False):
        return settings.notificationsTaskScheduler.action.add_item(message, notifier="discord", force_next=force)

    def _notify_discord(self, message="", force=False):
        if not settings.USE_DISCORD and not force:
            return False

        return self._send_discord(message, force=force)



from .anime import ConfigAnime
from .backup import ConfigBackupRestore
from .general import ConfigGeneral
from .index import Config
from .notifications import ConfigNotifications
from .post_processing import ConfigPostProcessing
from .providers import ConfigProviders
from .search import ConfigSearch
from .shares import ConfigShares
from .subtitles import ConfigSubtitles



# Author: Marvin Pinto <me@marvinp.ca>
# Author: Dennis Lutter <lad1337@gmail.com>
import time

import requests

from sickchill import logger, settings
from sickchill.oldbeard.common import (
    NOTIFY_DOWNLOAD,
    NOTIFY_GIT_UPDATE,
    NOTIFY_GIT_UPDATE_TEXT,
    NOTIFY_LOGIN,
    NOTIFY_LOGIN_TEXT,
    NOTIFY_SNATCH,
    NOTIFY_SUBTITLE_DOWNLOAD,
    notifyStrings,
)
from sickchill.oldbeard.helpers import make_session

API_URL = "https://api.pushover.net/1/messages.json"


class Notifier(object):
    def __init__(self):
        self.api_url = "https://api.pushover.net/1/messages.json"
        self.session = make_session()

    def test_notify(self, userKey=None, apiKey=None):
        return self._notify_pushover("This is a test notification from SickChill", "Test", userKey=userKey, apiKey=apiKey, force=True)

    def _send_pushover(self, msg, title, sound=None, userKey=None, apiKey=None, priority=None):
        """
        Sends a pushover notification to the address provided

        msg: The message to send (str)
        title: The title of the message
        sound: The notification sound to use
        userKey: The pushover user id to send the message to (or to subscribe with)
        apiKey: The pushover api key to use
        returns: True if the message succeeded, False otherwise
        """

        userKey = userKey or settings.PUSHOVER_USERKEY
        apiKey = apiKey or settings.PUSHOVER_APIKEY
        sound = sound or settings.PUSHOVER_SOUND
        priority = priority or settings.PUSHOVER_PRIORITY

        if not (userKey and apiKey):
            logger.warning("You must set a user key and api key to use pushover")
            return False

        logger.debug("Pushover API KEY in use: " + apiKey)

        try:
            data = {
                "token": apiKey,
                "user": userKey,
                "title": title,
                "message": msg.strip(),
                "timestamp": int(time.time()),
                "retry": 60,
                "expire": 3600,
                "priority": priority,
            }

            if sound and sound != "default":
                data["sound"] = sound

            if settings.PUSHOVER_DEVICE:
                data["device"] = settings.PUSHOVER_DEVICE

            response = self.session.post(self.api_url, data=data)
            response.raise_for_status()
        except requests.exceptions.RequestException as error:
            if error.response:
                print(error.response)

                # HTTP status 404 if the provided email address isn't a Pushover user.
                if error.response.status_code == 404:
                    logger.warning("Username is wrong/not a pushover email. Pushover will send an email to it")
                # For HTTP status code 401's, it is because you are passing in either an invalid token, or the user has not added your service.
                elif error.response.status_code == 401:
                    # HTTP status 401 if the user doesn't have the service added
                    subscribe_note = self._send_pushover(msg, title, sound=sound, userKey=userKey, apiKey=apiKey)
                    if subscribe_note:
                        logger.debug("Subscription sent")
                        return True
                    else:
                        logger.warning("Subscription could not be sent")
                # If you receive an HTTP status code of 400, it is because you failed to send the proper parameters
                elif error.response.status_code == 400:
                    logger.warning("Wrong data sent to pushover")

                # If you receive a HTTP status code of 429, it is because the message limit has been reached (free limit is 7,500)
                elif error.response.status_code == 429:
                    logger.warning("Pushover API message limit reached - try a different API key")

                return False

            logger.warning(f"Pushover notification failed. {error}")
            return False

        logger.info("Pushover notification successful.")
        return True

    def notify_snatch(self, ep_name, title=notifyStrings[NOTIFY_SNATCH]):
        if settings.PUSHOVER_NOTIFY_ONSNATCH:
            self._notify_pushover(title, ep_name)

    def notify_download(self, ep_name, title=notifyStrings[NOTIFY_DOWNLOAD]):
        if settings.PUSHOVER_NOTIFY_ONDOWNLOAD:
            self._notify_pushover(title, ep_name)

    def notify_subtitle_download(self, ep_name, lang, title=notifyStrings[NOTIFY_SUBTITLE_DOWNLOAD]):
        if settings.PUSHOVER_NOTIFY_ONSUBTITLEDOWNLOAD:
            self._notify_pushover(title, ep_name + ": " + lang)

    def notify_git_update(self, new_version="??"):
        if settings.USE_PUSHOVER:
            update_text = notifyStrings[NOTIFY_GIT_UPDATE_TEXT]
            title = notifyStrings[NOTIFY_GIT_UPDATE]
            self._notify_pushover(title, update_text + new_version)

    def notify_login(self, ipaddress=""):
        if settings.USE_PUSHOVER:
            update_text = notifyStrings[NOTIFY_LOGIN_TEXT]
            title = notifyStrings[NOTIFY_LOGIN]
            self._notify_pushover(title, update_text.format(ipaddress))

    def _notify_pushover(self, title, message, sound=None, userKey=None, apiKey=None, force=False):
        """
        Sends a pushover notification based on the provided info or SC config

        title: The title of the notification to send
        message: The message string to send
        sound: The notification sound to use
        userKey: The userKey to send the notification to
        apiKey: The apiKey to use to send the notification
        force: Enforce sending, for instance for testing
        """

        if not settings.USE_PUSHOVER and not force:
            logger.debug("Notification for Pushover not enabled, skipping this notification")
            return False

        logger.debug("Sending notification for " + message)

        return self._send_pushover(message, title, sound=sound, userKey=userKey, apiKey=apiKey)



import requests

from sickchill import logger, settings
from sickchill.oldbeard import helpers

meta_session = helpers.make_session()


def getShowImage(url, imgNum=None):
    if not url:
        return None

    # if they provided a fanart number try to use it instead
    if imgNum is not None:
        temp_url = url.split("-")[0] + "-" + str(imgNum) + ".jpg"
    else:
        temp_url = url

    logger.debug("Fetching image from " + temp_url)

    try:
        image_data = helpers.getURL(temp_url, session=meta_session, returns="content", allow_proxy=settings.PROXY_INDEXERS)
    except requests.exceptions.RequestException:
        image_data = None

    if not image_data:
        logger.warning("There was an error trying to retrieve the image, aborting")
        return

    return image_data



import base64
import datetime
import json
import os
import time
import traceback
from concurrent.futures import ThreadPoolExecutor
from mimetypes import guess_type
from operator import attrgetter
from secrets import compare_digest
from urllib.parse import urljoin

from mako.exceptions import RichTraceback
from tornado.concurrent import run_on_executor
from tornado.escape import utf8, xhtml_escape
from tornado.gen import coroutine
from tornado.process import cpu_count
from tornado.web import authenticated, HTTPError, RequestHandler

import sickchill.start
from sickchill import logger, settings
from sickchill.init_helpers import locale_dir
from sickchill.show.ComingEpisodes import ComingEpisodes
from sickchill.views.routes import Route

from ..oldbeard import db, helpers, network_timezones, ui
from .api.webapi import function_mapper
from .common import PageTemplate

try:
    import jwt
    from jwt.algorithms import RSAAlgorithm as jwt_algorithms_RSAAlgorithm

    has_cryptography = True
except Exception:
    has_cryptography = False


class BaseHandler(RequestHandler):
    def data_received(self, chunk):
        pass

    def __init__(self, *args, **kwargs):
        self.startTime = time.time()

        super().__init__(*args, **kwargs)
        # self.include_host = True

    # def set_default_headers(self):
    #     self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')

    def write_error(self, status_code, **kwargs):
        # handle 404 http errors
        if status_code == 404:
            url = self.request.uri
            if settings.WEB_ROOT and self.request.uri.startswith(settings.WEB_ROOT):
                url = url[len(settings.WEB_ROOT) + 1 :]

            if url[:3] != "api":
                t = PageTemplate(rh=self, filename="404.mako")
                return t.render(title="404", header=_("Oops: 404 Not Found"))
            else:
                return self.finish(_("Wrong API key used"))

        elif self.settings.get("debug") and "exc_info" in kwargs:
            exc_info = kwargs["exc_info"]
            trace_info = "".join(["{0}<br>".format(line) for line in traceback.format_exception(*exc_info)])
            request_info = "".join(["<strong>{0}</strong>: {1}<br>".format(k, self.request.__dict__[k]) for k in self.request.__dict__])
            error = exc_info[1]

            self.set_header("Content-Type", "text/html")
            return self.finish(
                """<html>
                                 <title>{0}</title>
                                 <body>
                                    <h2>Error</h2>
                                    <p>{1}</p>
                                    <h2>Traceback</h2>
                                    <p>{2}</p>
                                    <h2>Request Info</h2>
                                    <p>{3}</p>
                                    <button onclick="window.location='{4}/errorlogs/';">View Log(Errors)</button>
                                 </body>
                               </html>""".format(
                    error, error, trace_info, request_info, settings.WEB_ROOT
                )
            )

    def redirect(self, url, permanent=False, status=None):
        """Sends a redirect to the given (optionally relative) URL.

        ----->>>>> NOTE: Removed self.finish <<<<<-----

        If the ``status`` argument is specified, that value is used as the
        HTTP status code; otherwise either 301 (permanent) or 302
        (temporary) is chosen based on the ``permanent`` argument.
        The default is 302 (temporary).
        """
        if not url.startswith(settings.WEB_ROOT):
            url = settings.WEB_ROOT + url

        if self._headers_written:
            raise Exception("Cannot redirect after headers have been written")
        if not status:
            status = 301 if permanent else 302
        else:
            assert isinstance(status, int)
            assert 300 <= status <= 399
        self.set_status(status)
        self.set_header("Location", urljoin(utf8(self.request.uri), utf8(url)))

    def get_current_user(self):
        if isinstance(self, UI):
            return True

        if settings.WEB_USERNAME and settings.WEB_PASSWORD:
            # Authenticate using jwt for CF Access
            # NOTE: Setting a username and password is STILL required to protect poorly configured tunnels or firewalls
            if settings.CF_AUTH_DOMAIN and settings.CF_POLICY_AUD and has_cryptography:
                CERTS_URL = "{}/cdn-cgi/access/certs".format(settings.CF_AUTH_DOMAIN)
                if "CF_Authorization" in self.request.cookies:
                    jwk_set = helpers.getURL(CERTS_URL, returns="json")
                    for key_dict in jwk_set["keys"]:
                        public_key = jwt_algorithms_RSAAlgorithm.from_jwk(json.dumps(key_dict))
                        if jwt.decode(self.request.cookies["CF_Authorization"], key=public_key, audience=settings.CF_POLICY_AUD):
                            return True

            # Logged into UI?
            if self.get_secure_cookie("sickchill_user"):
                return True

            # Basic Auth at a minimum
            auth_header = self.request.headers.get("Authorization")
            if auth_header and auth_header.startswith("Basic "):
                auth_decoded = base64.decodebytes(auth_header[6:].encode()).decode()
                username, password = auth_decoded.split(":", 2)
                if compare_digest(username, settings.WEB_USERNAME) and compare_digest(password, settings.WEB_PASSWORD):
                    return True
                return False

        else:
            # Local network
            # strip <scope id> / <zone id> (%value/if_name) from remote_ip IPv6 scoped literal IP Addresses (RFC 4007) until phihag/ipaddress is updated tracking cpython 3.9.
            return helpers.is_ip_local(self.request.remote_ip.rsplit("%")[0])

    def get_user_locale(self):
        return settings.GUI_LANG or None


class WebHandler(BaseHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.executor = ThreadPoolExecutor(cpu_count(), thread_name_prefix="WEBSERVER-" + self.__class__.__name__.upper())

    @authenticated
    @coroutine
    def get(self, route, *args, **kwargs):
        try:
            # route -> method obj
            route = route.strip("/").replace(".", "_").replace("-", "_") or "index"
            method = getattr(self, route)

            results = yield self.async_call(method)

            self.finish(results)

        except AttributeError:
            logger.debug('Failed doing webui request "{0}": {1}'.format(route, traceback.format_exc()))
            raise HTTPError(404)

    @run_on_executor
    def async_call(self, function):
        try:
            # TODO: Make all routes use get_argument so we can take advantage of tornado's argument sanitization, separate post and get, and get rid of this
            # nonsense loop so we can just yield the method directly
            # raise Exception('Raising from async_call')
            kwargs = self.request.arguments
            for arg, value in kwargs.items():
                if len(value) == 1:
                    kwargs[arg] = xhtml_escape(value[0])
                elif isinstance(value, str):
                    kwargs[arg] = xhtml_escape(value)
                elif isinstance(value, list):
                    kwargs[arg] = [xhtml_escape(v) for v in value]
                else:
                    raise Exception
            return function(**kwargs)
        except TypeError:
            return function()
        except Exception as error:
            return WebRoot(application=self.application, request=self.request).print_traceback(error=error, **self.request.arguments)

    # post uses get method
    post = get


@Route("(.*)(/?)", name="index")
class WebRoot(WebHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def print_traceback(self, error, *args, **kwargs):
        logger.info(f"A mako error occurred: {error}")
        t = PageTemplate(rh=self, filename="500.mako")
        kwargs["backtrace"] = RichTraceback(error=error)
        return t.render(*args, **kwargs)

    def index(self):
        return self.redirect("/" + settings.DEFAULT_PAGE + "/")

    def robots_txt(self):
        """ Keep web crawlers out """
        self.set_header("Content-Type", "text/plain")
        return "User-agent: *\nDisallow: /"

    def apibuilder(self):
        main_db_con = db.DBConnection(row_type="dict")
        shows = sorted(settings.showList, key=lambda mbr: attrgetter("sort_name")(mbr))
        episodes = {}

        results = main_db_con.select("SELECT episode, season, showid " "FROM tv_episodes " "ORDER BY season ASC, episode ASC")

        for result in results:
            if result["showid"] not in episodes:
                episodes[result["showid"]] = {}

            if result["season"] not in episodes[result["showid"]]:
                episodes[result["showid"]][result["season"]] = []

            episodes[result["showid"]][result["season"]].append(result["episode"])

        if len(settings.API_KEY) == 32:
            apikey = settings.API_KEY
        else:
            apikey = _("API Key not generated")

        t = PageTemplate(rh=self, filename="apiBuilder.mako")
        return t.render(title=_("API Builder"), header=_("API Builder"), shows=shows, episodes=episodes, apikey=apikey, commands=function_mapper)

    def setHomeLayout(self):
        layout = self.get_query_argument("layout")
        if layout not in ("poster", "small", "banner", "simple", "coverflow"):
            layout = "poster"

        settings.HOME_LAYOUT = layout
        # Don't redirect to default page so user can see new layout
        return self.redirect("/home/")

    def setPosterSortBy(self):
        sort = self.get_query_argument("sort")
        if sort not in ("name", "date", "network", "progress"):
            sort = "name"

        settings.POSTER_SORTBY = sort
        sickchill.start.save_config()

    def setPosterSortDir(self):
        direction = self.get_query_argument("direction")

        settings.POSTER_SORTDIR = int(direction)
        sickchill.start.save_config()

    def setHistoryLayout(self):
        layout = self.get_query_argument("layout")
        if layout not in ("compact", "detailed"):
            layout = "detailed"

        settings.HISTORY_LAYOUT = layout

        return self.redirect("/history/")

    def toggleDisplayShowSpecials(self):
        show = self.get_query_argument("show")
        settings.DISPLAY_SHOW_SPECIALS = not settings.DISPLAY_SHOW_SPECIALS

        return self.redirect("/home/displayShow?show=" + show)

    def setScheduleLayout(self):
        layout = self.get_query_argument("layout")
        if layout not in ("poster", "banner", "list", "calendar"):
            layout = "banner"

        if layout == "calendar":
            settings.COMING_EPS_SORT = "date"

        settings.COMING_EPS_LAYOUT = layout

        return self.redirect("/schedule/")

    def toggleScheduleDisplayPaused(self):

        settings.COMING_EPS_DISPLAY_PAUSED = not settings.COMING_EPS_DISPLAY_PAUSED

        return self.redirect("/schedule/")

    def toggleScheduleDisplaySnatched(self):

        settings.COMING_EPS_DISPLAY_SNATCHED = not settings.COMING_EPS_DISPLAY_SNATCHED

        return self.redirect("/schedule/")

    def setScheduleSort(self):
        sort = self.get_query_argument("sort")
        if sort not in ("date", "network", "show") or settings.COMING_EPS_LAYOUT == "calendar":
            sort = "date"

        settings.COMING_EPS_SORT = sort

        return self.redirect("/schedule/")

    def schedule(self):
        layout = self.get_query_argument("layout", settings.COMING_EPS_LAYOUT)
        next_week = datetime.date.today() + datetime.timedelta(days=7)
        next_week1 = datetime.datetime.combine(next_week, datetime.time(tzinfo=network_timezones.sb_timezone))
        results = ComingEpisodes.get_coming_episodes(ComingEpisodes.categories, settings.COMING_EPS_SORT, False)
        today = datetime.datetime.now().replace(tzinfo=network_timezones.sb_timezone)

        # Allow local overriding of layout parameter
        if layout not in ("poster", "banner", "list", "calendar"):
            layout = settings.COMING_EPS_LAYOUT

        t = PageTemplate(rh=self, filename="schedule.mako")
        return t.render(
            next_week=next_week1,
            today=today,
            results=results,
            layout=layout,
            title=_("Schedule"),
            header=_("Schedule"),
            topmenu="schedule",
            controller="schedule",
            action="index",
        )


@Route("/ui(/?.*)", name="ui")
class UI(WebRoot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def locale_json(self):

        lang = self.get_query_argument("lang")
        """ Get /locale/{lang_code}/LC_MESSAGES/messages.json """
        locale_file = os.path.normpath(f"{locale_dir()}/{lang}/LC_MESSAGES/messages.json")

        if os.path.isfile(locale_file):
            self.set_header("Content-Type", "application/json")
            with open(locale_file) as content:
                return content.read()
        else:
            self.set_status(204)  # "No Content"
            return None

    @staticmethod
    def add_message():
        ui.notifications.message(_("Test 1"), _("This is test number 1"))
        ui.notifications.error(_("Test 2"), _("This is test number 2"))

        return "ok"

    def get_messages(self):
        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        self.set_header("Content-Type", "application/json")
        notifications = ui.notifications.get_notifications(self.request.remote_ip)
        messages = {}
        for index, cur_notification in enumerate(notifications, 1):
            messages["notification-" + str(index)] = {
                "hash": hash(cur_notification),
                "title": cur_notification.title,
                "message": cur_notification.message,
                "type": cur_notification.type,
            }

        return json.dumps(messages)

    def set_site_message(self):
        message = self.get_body_argument("message", None)
        tag = self.get_body_argument("tag", None)
        level = self.get_body_argument("level")
        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        if message:
            helpers.add_site_message(message, tag=tag, level=level)
        else:
            if settings.BRANCH and self.get_current_user() and settings.BRANCH not in ("master", "pip") and not settings.DEVELOPER:
                message = _("You're using the {branch} branch. Please use 'master' unless specifically asked".format(branch=settings.BRANCH))
                helpers.add_site_message(message, tag="not_using_master_branch", level="danger")

        return settings.SITE_MESSAGES

    def get_site_messages(self):
        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        return settings.SITE_MESSAGES

    def dismiss_site_message(self):
        index = self.get_query_argument("index")
        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        helpers.remove_site_message(key=index)
        return settings.SITE_MESSAGES

    def sickchill_background(self):
        if settings.SICKCHILL_BACKGROUND_PATH and os.path.isfile(settings.SICKCHILL_BACKGROUND_PATH):
            self.set_header("Content-Type", guess_type(settings.SICKCHILL_BACKGROUND_PATH)[0])
            with open(settings.SICKCHILL_BACKGROUND_PATH, "rb") as content:
                return content.read()
        return None

    def custom_css(self):
        if settings.CUSTOM_CSS_PATH and os.path.isfile(settings.CUSTOM_CSS_PATH):
            self.set_header("Content-Type", "text/css")
            with open(settings.CUSTOM_CSS_PATH, "r") as content:
                return content.read()
        return None



import re
from urllib.parse import urljoin

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("TorrentBytes")

        # Credentials
        self.username = None
        self.password = None

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0
        self.freeleech = False

        # URLs
        self.url = "https://www.torrentbytes.net"
        self.urls = {"login": urljoin(self.url, "takelogin.php"), "search": urljoin(self.url, "browse.php")}

        # Proper Strings
        self.proper_strings = ["PROPER", "REPACK"]

        # Cache
        self.cache = tvcache.TVCache(self)

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {"username": self.username, "password": self.password, "login": "Log in!"}

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("Username or password incorrect", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        search_params = {"Episode": {"c33": 1, "c38": 1, "c32": 1, "c37": 1}, "Season": {"c41": 1}}

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:

                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                search_params[mode]["search"] = search_string
                data = self.get_url(self.urls["search"], params=search_params[mode], returns="text")
                if not data:
                    logger.debug("No data returned from provider")
                    continue

                with BS4Parser(data, "html5lib") as html:
                    torrent_table = html.find("table", border="1")
                    torrent_rows = torrent_table("tr") if torrent_table else []

                    # Continue only if at least one Release is found
                    if len(torrent_rows) < 2:
                        logger.debug("Data returned from provider does not contain any torrents")
                        continue

                    # "Type", "Name", Files", "Comm.", "Added", "TTL", "Size", "Snatched", "Seeders", "Leechers"
                    labels = [label.get_text(strip=True) for label in torrent_rows[0]("td")]

                    for result in torrent_rows[1:]:
                        try:
                            cells = result("td")

                            download_url = urljoin(self.url, cells[labels.index("Name")].find("a", href=re.compile(r"download.php\?id="))["href"])
                            title_element = cells[labels.index("Name")].find("a", href=re.compile(r"details.php\?id="))
                            title = title_element.get("title", "") or title_element.get_text(strip=True)
                            if not all([title, download_url]):
                                continue

                            if self.freeleech:
                                # Free leech torrents are marked with green [F L] in the title (i.e. <font color=green>[F&nbsp;L]</font>)
                                freeleech = cells[labels.index("Name")].find("font", color="green")
                                if not freeleech or freeleech.get_text(strip=True) != "[F\xa0L]":
                                    continue

                            seeders = try_int(cells[labels.index("Seeders")].get_text(strip=True))
                            leechers = try_int(cells[labels.index("Leechers")].get_text(strip=True))

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            # Need size for failed downloads handling
                            torrent_size = cells[labels.index("Size")].get_text(strip=True)
                            size = convert_size(torrent_size) or -1
                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}

                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)
                        except (AttributeError, TypeError, ValueError):
                            continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



from bs4 import BeautifulSoup


class BS4Parser(object):
    def __init__(self, *args, **kwargs):
        self.soup = BeautifulSoup(*args, **kwargs)

    def __enter__(self) -> BeautifulSoup:
        return self.soup

    def __exit__(self, exc_ty, exc_val, tb):
        self.soup.clear(True)
        self.soup = None



import markdown2

import sickchill.start
from sickchill import logger, settings

from .common import PageTemplate
from .home import Home
from .routes import Route


@Route("/news(/?.*)", name="news")
class HomeNews(Home):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def index(self):
        # noinspection PyBroadException
        try:
            news = settings.versionCheckScheduler.action.check_for_new_news()
        except Exception:
            logger.debug("Could not load news from repo, giving a link!")
            news = _("Could not load news from the repo. [Click here for news.md])({news_url})").format(news_url=settings.NEWS_URL)

        settings.NEWS_LAST_READ = settings.NEWS_LATEST
        settings.NEWS_UNREAD = 0
        sickchill.start.save_config()

        t = PageTemplate(rh=self, filename="markdown.mako")
        data = markdown2.markdown(news if news else _("The was a problem connecting to github, please refresh and try again"), extras=["header-ids"])

        return t.render(title=_("News"), header=_("News"), topmenu="system", data=data, controller="news", action="index")



from datetime import datetime

import bencodepy
from feedparser import FeedParserDict

from sickchill import logger, settings
from sickchill.helper.common import try_int
from sickchill.oldbeard.classes import Proper, TorrentSearchResult
from sickchill.oldbeard.common import Quality
from sickchill.oldbeard.db import DBConnection
from sickchill.providers.GenericProvider import GenericProvider
from sickchill.show.Show import Show


class TorrentProvider(GenericProvider):
    def __init__(self, name):
        super().__init__(name)
        self.ratio = None
        self.provider_type = GenericProvider.TORRENT

    def find_propers(self, search_date=None):
        results = []
        db = DBConnection()
        placeholders = ", ".join(["?"] * len(Quality.DOWNLOADED + Quality.SNATCHED + Quality.SNATCHED_BEST))
        sql_results = db.select(
            f"SELECT s.show_name, e.showid, e.season, e.episode, e.status, e.airdate FROM tv_episodes AS e INNER JOIN tv_shows AS s ON (e.showid = s.indexer_id) WHERE e.airdate >= ? AND e.status IN ({placeholders}) and e.is_proper = 0",
            [search_date.toordinal(), *Quality.DOWNLOADED + Quality.SNATCHED + Quality.SNATCHED_BEST],
        )

        for result in sql_results or []:
            show = Show.find(settings.showList, int(result["showid"]))

            if show:
                episode = show.getEpisode(result["season"], result["episode"])

                for term in self.proper_strings:
                    search_strings = self.get_episode_search_strings(episode, add_string=term)

                    for search_string in search_strings:
                        for item in self.search(search_string):
                            title, url = self._get_title_and_url(item)

                            results.append(Proper(title, url, datetime.today(), show))

        return results

    @property
    def is_active(self):
        return bool(settings.USE_TORRENTS) and self.is_enabled

    @property
    def _custom_trackers(self):
        if not (settings.TRACKERS_LIST and self.public):
            return ""

        return "&tr=" + "&tr=".join({x.strip() for x in settings.TRACKERS_LIST.split(",") if x.strip()})

    def _get_result(self, episodes):
        return TorrentSearchResult(episodes)

    def _get_size(self, item):
        if isinstance(item, dict):
            size = item.get("size", -1)
        elif isinstance(item, (list, tuple)) and len(item) > 2:
            size = item[2]
        else:
            size = -1

        size = try_int(size, -1)

        # Make sure we didn't select seeds/leechers by accident
        if not size or size < 1024 * 1024:
            size = -1

        return size

    def _get_storage_dir(self):
        return settings.TORRENT_DIR

    def _get_title_and_url(self, item):
        if isinstance(item, (dict, FeedParserDict)):
            download_url = item.get("url", "")
            title = item.get("title", "")

            if not download_url:
                download_url = item.get("link", "")
        elif isinstance(item, (list, tuple)) and len(item) > 1:
            download_url = item[1]
            title = item[0]
        else:
            download_url = ""
            title = ""

        if title.endswith("DIAMOND"):
            logger.info("Skipping DIAMOND release for mass fake releases.")
            download_url = title = "FAKERELEASE"

        if download_url:
            download_url = download_url.replace("&amp;", "&")

        if title:
            title = title.replace(" ", ".")

        return title, download_url

    def _verify_download(self, filename):
        try:
            bencodepy.bread(filename)
        except bencodepy.BencodeDecodeError as e:
            logger.debug("Failed to validate torrent file: {0}".format(str(e)))
            logger.debug("Result is not a valid torrent file")
            return False
        return True

    def seed_ratio(self):
        return self.ratio



import hashlib
import os
import time
from functools import reduce
from pathlib import Path
from xml.etree import ElementTree

# http://www.radicand.org/blog/orz/2010/2/21/edonkey2000-hash-in-python/
import requests


def get_file_hash(filePath: Path):
    """ Returns the ed2k hash of a given file."""

    md4 = hashlib.new("md4").copy

    def gen(f):
        while True:
            x = f.read(9728000)
            if x:
                yield x
            else:
                return

    def md4_hash(data):
        m = md4()
        m.update(data)
        return m

    with filePath.open("rb") as f:
        a = gen(f)
        hashes = [md4_hash(data).digest() for data in a]
        if len(hashes) == 1:
            return hashes[0].hex()
        else:
            return md4_hash(reduce(lambda b, c: b + c, hashes, b"")).hexdigest()


def download_file(url, filename: Path):
    try:
        r = requests.get(url, stream=True, verify=False)
        r.raise_for_status()
        with filename.open("wb") as fp:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    fp.write(chunk)
                    fp.flush()

    except requests.exceptions.RequestException:
        filename.unlink(True)
        return False

    return True


def get_anime_titles_xml(path: Path):
    return download_file("https://raw.githubusercontent.com/ScudLee/anime-lists/master/animetitles.xml", path)


def get_anime_list_xml(path: Path):
    return download_file("https://raw.githubusercontent.com/ScudLee/anime-lists/master/anime-list.xml", path)


def read_anidb_xml(cache_dir: Path):
    file_path = cache_dir / "animetitles.xml"

    if not file_path.exists():
        if not get_anime_titles_xml(file_path):
            return
    else:
        mtime = os.path.getmtime(file_path)
        if time.time() > mtime + 24 * 60 * 60:
            if not get_anime_titles_xml(file_path):
                return

    return read_xml_into_etree(file_path)


def read_tvdb_map_xml(cache_dir: Path):
    file_path = cache_dir / "anime-list.xml"
    if not file_path.is_file():
        if not get_anime_list_xml(file_path):
            return
    else:
        mtime = os.path.getmtime(file_path)
        if time.time() > mtime + 24 * 60 * 60:
            if not get_anime_list_xml(file_path):
                return

    return read_xml_into_etree(file_path)


def read_xml_into_etree(file_path):
    with file_path.open("r") as f:
        return ElementTree.ElementTree(file=f)



import operator
import platform
import re
import uuid
from functools import reduce
from os import path

from sickchill import settings, version
from sickchill.helper import video_screen_size
from sickchill.init_helpers import setup_gettext
from sickchill.recompiled import tags
from sickchill.tagger.episode import EpisodeTags

from .numdict import NumDict

setup_gettext()

# If some provider has an issue with functionality of SC, other than user agents, it's best to come talk to us rather than block.
# It is no different than us going to a provider if we have questions or issues. Be a team player here.
# This is disabled, was only added for testing, and has no config.ini or web ui setting. To enable, set SPOOF_USER_AGENT = True
INSTANCE_ID = str(uuid.uuid1())
USER_AGENT = "SickChill/{version} ({os} {architecture} {os_version}; {instance})".format(
    version=version.__version__, os=platform.system(), architecture=platform.machine(), os_version=platform.release(), instance=INSTANCE_ID
)

cpu_presets = {"HIGH": 5, "NORMAL": 2, "LOW": 1}

# Other constants
MULTI_EP_RESULT = -1
SEASON_RESULT = -2

# Notification Types
NOTIFY_SNATCH = 1
NOTIFY_DOWNLOAD = 2
NOTIFY_SUBTITLE_DOWNLOAD = 3
NOTIFY_GIT_UPDATE = 4
NOTIFY_GIT_UPDATE_TEXT = 5
NOTIFY_LOGIN = 6
NOTIFY_LOGIN_TEXT = 7
NOTIFY_POSTPROCESS = 8

notifyStrings = NumDict(
    {
        NOTIFY_SNATCH: _("Started Download"),
        NOTIFY_DOWNLOAD: _("Finished Download"),
        NOTIFY_SUBTITLE_DOWNLOAD: _("Subtitle Download Finished"),
        NOTIFY_GIT_UPDATE: _("SickChill Updated"),
        NOTIFY_GIT_UPDATE_TEXT: _("SickChill Updated To Commit#: "),
        NOTIFY_LOGIN: _("SickChill new login"),
        NOTIFY_LOGIN_TEXT: _("New login from IP: {0}. http://geomaplookup.net/?ip={0}"),
        NOTIFY_POSTPROCESS: _("Finished Post Processing"),
    }
)

# Episode statuses
UNKNOWN = -1  # should never happen
UNAIRED = 1  # episodes that haven't aired yet
SNATCHED = 2  # qualified with quality
WANTED = 3  # episodes we don't have but want to get
DOWNLOADED = 4  # qualified with quality
SKIPPED = 5  # episodes we don't want
ARCHIVED = 6  # episodes that you don't have locally (counts toward download completion stats)
IGNORED = 7  # episodes that you don't want included in your download stats
SNATCHED_PROPER = 9  # qualified with quality
SUBTITLED = 10  # qualified with quality
FAILED = 11  # episode downloaded or snatched we don't want
SNATCHED_BEST = 12  # episode re-downloaded using best quality

NAMING_REPEAT = 1
NAMING_EXTEND = 2
NAMING_DUPLICATE = 4
NAMING_LIMITED_EXTEND = 8
NAMING_SEPARATED_REPEAT = 16
NAMING_LIMITED_EXTEND_E_PREFIXED = 32

MULTI_EP_STRINGS = NumDict(
    {
        NAMING_REPEAT: _("Repeat"),
        NAMING_SEPARATED_REPEAT: _("Repeat (Separated)"),
        NAMING_DUPLICATE: _("Duplicate"),
        NAMING_EXTEND: _("Extend"),
        NAMING_LIMITED_EXTEND: _("Extend (Limited)"),
        NAMING_LIMITED_EXTEND_E_PREFIXED: _("Extend (Limited, E-prefixed)"),
    }
)


class Quality(object):
    """
    Determine quality and set status codes
    """

    NONE = 0  # 0
    SDTV = 1  # 1
    SDDVD = 1 << 1  # 2
    HDTV = 1 << 2  # 4
    RAWHDTV = 1 << 3  # 8  -- 720p/1080i mpeg2 (trollhd releases)
    FULLHDTV = 1 << 4  # 16 -- 1080p HDTV (QCF releases)
    HDWEBDL = 1 << 5  # 32
    FULLHDWEBDL = 1 << 6  # 64 -- 1080p web-dl
    HDBLURAY = 1 << 7  # 128
    FULLHDBLURAY = 1 << 8  # 256
    UHD_4K_TV = 1 << 9  # 512 -- 2160p aka 4K UHD aka UHD-1
    UHD_4K_WEBDL = 1 << 10  # 1024
    UHD_4K_BLURAY = 1 << 11  # 2048
    UHD_8K_TV = 1 << 12  # 4096 -- 4320p aka 8K UHD aka UHD-2
    UHD_8K_WEBDL = 1 << 13  # 8192
    UHD_8K_BLURAY = 1 << 14  # 16384
    ANYHDTV = HDTV | FULLHDTV  # 20
    ANYWEBDL = HDWEBDL | FULLHDWEBDL  # 96
    ANYBLURAY = HDBLURAY | FULLHDBLURAY  # 384

    # put these bits at the other end of the spectrum, far enough out that they shouldn't interfere
    UNKNOWN = 1 << 15  # 32768

    qualityStrings = NumDict(
        {
            # None: "None",
            NONE: "N/A",
            UNKNOWN: "Unknown",
            SDTV: "SDTV",
            SDDVD: "SD DVD",
            HDTV: "720p HDTV",
            RAWHDTV: "RawHD",
            FULLHDTV: "1080p HDTV",
            HDWEBDL: "720p WEB-DL",
            FULLHDWEBDL: "1080p WEB-DL",
            HDBLURAY: "720p BluRay",
            FULLHDBLURAY: "1080p BluRay",
            UHD_4K_TV: "4K UHD TV",
            UHD_8K_TV: "8K UHD TV",
            UHD_4K_WEBDL: "4K UHD WEB-DL",
            UHD_8K_WEBDL: "8K UHD WEB-DL",
            UHD_4K_BLURAY: "4K UHD BluRay",
            UHD_8K_BLURAY: "8K UHD BluRay",
        }
    )

    sceneQualityStrings = NumDict(
        {
            # None: "None",
            NONE: "N/A",
            UNKNOWN: "Unknown",
            SDTV: "HDTV",
            SDDVD: "",
            HDTV: "720p HDTV",
            RAWHDTV: "1080i HDTV",
            FULLHDTV: "1080p HDTV",
            HDWEBDL: "720p WEB-DL",
            FULLHDWEBDL: "1080p WEB-DL",
            HDBLURAY: "720p BluRay",
            FULLHDBLURAY: "1080p BluRay",
            UHD_4K_TV: "4K UHD TV",
            UHD_8K_TV: "8K UHD TV",
            UHD_4K_WEBDL: "4K UHD WEB-DL",
            UHD_8K_WEBDL: "8K UHD WEB-DL",
            UHD_4K_BLURAY: "4K UHD BluRay",
            UHD_8K_BLURAY: "8K UHD BluRay",
        }
    )

    combinedQualityStrings = NumDict({ANYHDTV: "HDTV", ANYWEBDL: "WEB-DL", ANYBLURAY: "BluRay"})

    cssClassStrings = NumDict(
        {
            # None: "None",
            NONE: "N/A",
            UNKNOWN: "Unknown",
            SDTV: "SDTV",
            SDDVD: "SDDVD",
            HDTV: "HD720p",
            RAWHDTV: "RawHD",
            FULLHDTV: "HD1080p",
            HDWEBDL: "HD720p",
            FULLHDWEBDL: "HD1080p",
            HDBLURAY: "HD720p",
            FULLHDBLURAY: "HD1080p",
            UHD_4K_TV: "UHD-4K",
            UHD_8K_TV: "UHD-8K",
            UHD_4K_WEBDL: "UHD-4K",
            UHD_8K_WEBDL: "UHD-8K",
            UHD_4K_BLURAY: "UHD-4K",
            UHD_8K_BLURAY: "UHD-8K",
            ANYHDTV: "any-hd",
            ANYWEBDL: "any-hd",
            ANYBLURAY: "any-hd",
        }
    )

    statusPrefixes = NumDict(
        {
            DOWNLOADED: _("Downloaded"),
            SNATCHED: _("Snatched"),
            SNATCHED_PROPER: _("Snatched (Proper)"),
            FAILED: _("Failed"),
            SNATCHED_BEST: _("Snatched (Best)"),
            ARCHIVED: _("Archived"),
        }
    )

    @staticmethod
    def _getStatusStrings(status):
        """
        Returns string values associated with Status prefix

        :param status: Status prefix to resolve
        :return: Human readable status value
        """
        to_return = {}
        for quality in Quality.qualityStrings:
            if quality is not None:
                stat = Quality.statusPrefixes[status]
                qual = Quality.qualityStrings[quality]
                comp = Quality.compositeStatus(status, quality)
                to_return[comp] = "{0} ({1})".format(stat, qual)
        return to_return

    @staticmethod
    def combineQualities(allowed_qualities, preferred_qualities):
        any_quality = 0
        best_quality = 0
        if allowed_qualities:
            any_quality = reduce(operator.or_, allowed_qualities)
        if preferred_qualities:
            best_quality = reduce(operator.or_, preferred_qualities)
        return any_quality | (best_quality << 16)

    @staticmethod
    def splitQuality(quality):
        if quality is None:
            quality = Quality.NONE
        allowed_qualities = []
        preferred_qualities = []
        for cur_qual in Quality.qualityStrings:
            if cur_qual is None:
                cur_qual = Quality.NONE
            if cur_qual & quality:
                allowed_qualities.append(cur_qual)
            if cur_qual << 16 & quality:
                preferred_qualities.append(cur_qual)

        return sorted(allowed_qualities), sorted(preferred_qualities)

    @staticmethod
    def nameQuality(name, anime=False):
        """
        Return The quality from an episode File renamed by SickChill
        If no quality is achieved it will try scene_quality regex

        :param name: to parse
        :param anime: Boolean to indicate if the show we're resolving is Anime
        :return: Quality prefix
        """

        # Try Scene names first
        quality = Quality.scene_quality(name, anime)
        if quality != Quality.UNKNOWN:
            return quality

        quality = Quality.qualityFromFileMeta(name)
        if quality != Quality.UNKNOWN:
            return quality

        if name.lower().endswith(".ts"):
            return Quality.RAWHDTV
        else:
            return Quality.UNKNOWN

    @staticmethod
    def scene_quality(name, anime=False):
        """
        Return The quality from the scene episode File

        :param name: Episode filename to analyse
        :param anime: Boolean to indicate if the show we're resolving is Anime
        :return: Quality
        """

        if not name:
            return Quality.UNKNOWN

        name = path.basename(name)

        result = None
        ep = EpisodeTags(name)

        if anime:
            sd_options = tags.anime_sd.search(name)
            hd_options = tags.anime_hd.search(name)
            full_hd = tags.anime_fullhd.search(name)
            ep.rex["bluray"] = tags.anime_bluray

            # BluRay
            if ep.bluray and (full_hd or hd_options):
                result = Quality.FULLHDBLURAY if full_hd else Quality.HDBLURAY
            # HD TV
            elif not ep.bluray and (full_hd or hd_options):
                result = Quality.FULLHDTV if full_hd else Quality.HDTV
            # SD DVD
            elif ep.dvd:
                result = Quality.SDDVD
            # SD TV
            elif sd_options:
                result = Quality.SDTV
        elif ep.hevc and not settings.QUALITY_ALLOW_HEVC:
            result = Quality.NONE
        elif ep.mpeg:
            result = Quality.RAWHDTV
        # Is it UHD?
        elif ep.vres in {2160, 4320} and ep.scan == "p":
            # BluRay
            full_res = ep.vres == 4320
            if ep.bluray:
                result = (Quality.UHD_4K_BLURAY, Quality.UHD_8K_BLURAY)[full_res]
            # WEB-DL
            elif ep.itunes or ep.amazon or ep.netflix or ep.web:
                result = (Quality.UHD_4K_WEBDL, Quality.UHD_8K_WEBDL)[full_res]
            # HDTV
            elif ep.tv == "hd":
                result = (Quality.UHD_4K_TV, Quality.UHD_8K_TV)[full_res]
        elif ep.vres in {1080, 720}:
            if ep.scan == "p":
                # BluRay
                full_res = ep.vres == 1080
                if ep.bluray or ep.hddvd:
                    result = (Quality.HDBLURAY, Quality.FULLHDBLURAY)[full_res]
                # WEB-DL
                elif ep.itunes or ep.amazon or ep.netflix or ep.web:
                    result = (Quality.HDWEBDL, Quality.FULLHDWEBDL)[full_res]
                # HDTV
                elif ep.tv == "hd" or ep.hevc:
                    result = (Quality.HDTV, Quality.FULLHDTV)[full_res]  # 1080 HDTV h264
                # MPEG2 encoded
                elif all([full_res, ep.tv == "hd", ep.mpeg]):
                    result = Quality.RAWHDTV
                elif all([not full_res, ep.tv == "hd", ep.mpeg]):
                    result = Quality.RAWHDTV
            elif (ep.res == "1080i") and ep.tv == "hd" and (ep.mpeg or (ep.raw and ep.avc_non_free)):
                result = Quality.RAWHDTV
        elif not ep.vres and ep.netflix or ep.amazon or ep.itunes:
            result = Quality.HDWEBDL
        elif ep.hrws:
            result = Quality.HDTV

        # Is it SD?
        elif ep.xvid or ep.avc:
            # SD DVD
            if ep.dvd or ep.bluray:
                result = Quality.SDDVD
            # SDTV
            elif ep.res == "480p" or any([ep.tv, ep.sat, ep.web]):
                result = Quality.SDTV
        elif ep.dvd:
            # SD DVD
            result = Quality.SDDVD
        elif ep.tv:
            # SD TV/HD TV
            result = (Quality.SDTV, Quality.HDTV)[ep.tv == "hd"]
        elif ep.raw or ep.mpeg:
            # RawHD
            result = Quality.RAWHDTV

        return Quality.UNKNOWN if result is None else result

    @staticmethod
    def qualityFromFileMeta(filename):
        """
        Get quality file file metadata

        :param filename: Filename to analyse
        :return: Quality prefix
        """

        height = video_screen_size(filename)[1]
        if not height:
            return Quality.UNKNOWN

        base_filename = path.basename(filename)
        bluray = re.search(r"blue?-?ray|hddvd|b[rd](rip|mux)", base_filename, re.I) is not None
        webdl = re.search(r"web.?dl|web(rip|mux|hd)", base_filename, re.I) is not None

        ret = Quality.UNKNOWN
        if 3240 < height:
            ret = ((Quality.UHD_8K_TV, Quality.UHD_8K_BLURAY)[bluray], Quality.UHD_8K_WEBDL)[webdl]
        if 1620 < height <= 3240:
            ret = ((Quality.UHD_4K_TV, Quality.UHD_4K_BLURAY)[bluray], Quality.UHD_4K_WEBDL)[webdl]
        elif 800 < height <= 1620:
            ret = ((Quality.FULLHDTV, Quality.FULLHDBLURAY)[bluray], Quality.FULLHDWEBDL)[webdl]
        elif 680 < height <= 800:
            ret = ((Quality.HDTV, Quality.HDBLURAY)[bluray], Quality.HDWEBDL)[webdl]
        elif height <= 680:
            ret = (Quality.SDTV, Quality.SDDVD)[re.search(r"dvd|b[rd]rip|blue?-?ray", base_filename, re.I) is not None]

        return ret

    @staticmethod
    def compositeStatus(status, quality):
        if quality is None:
            quality = Quality.NONE
        return status + 100 * quality

    @staticmethod
    def qualityDownloaded(status):
        return (status - DOWNLOADED) / 100

    @staticmethod
    def splitCompositeStatus(status):
        """
        Split a composite status code into a status and quality.

        :param status: to split
        :returns: a tuple containing (status, quality)
        """
        status = int(status)
        if status == UNKNOWN:
            return UNKNOWN, Quality.UNKNOWN

        for q in sorted(Quality.qualityStrings, reverse=True):
            if status > q * 100:
                return status - q * 100, q

        return status, Quality.NONE

    @staticmethod
    def sceneQualityFromName(name, quality):
        """
        Get scene naming parameters from filename and quality

        :param name: Filename to check
        :param quality: int of quality to make sure we get the right rip type
        :return: encoder type for scene quality naming
        """
        codec_list = ["xvid", "divx"]
        x264_list = ["x264", "x 264", "x.264"]
        h264_list = ["h264", "h 264", "h.264", "avc"]
        x265_list = ["x265", "x 265", "x.265"]
        h265_list = ["h265", "h 265", "h.265", "hevc"]
        codec_list += x264_list + h264_list + x265_list + h265_list

        found_codecs = {}
        found_codec = None

        for codec in codec_list:
            if codec in name.lower():
                found_codecs[name.lower().rfind(codec)] = codec

        if found_codecs:
            sorted_codecs = sorted(found_codecs, reverse=True)
            found_codec = found_codecs[list(sorted_codecs)[0]]

        # 2 corresponds to SDDVD quality
        if quality == 2:
            if re.search(r"b(r|d|rd)?(-| |\.)?(rip|mux)", name.lower()):
                rip_type = " BDRip"
            elif re.search(r"(dvd)(-| |\.)?(rip|mux)?", name.lower()):
                rip_type = " DVDRip"
            else:
                rip_type = ""

        if found_codec:
            if codec_list[0] in found_codec:
                found_codec = "XviD"
            elif codec_list[1] in found_codec:
                found_codec = "DivX"
            elif found_codec in x264_list:
                found_codec = x264_list[0]
            elif found_codec in h264_list:
                found_codec = h264_list[0]
            elif found_codec in x265_list:
                found_codec = x265_list[0]
            elif found_codec in h265_list:
                found_codec = h265_list[0]

            if quality == 2:
                return rip_type + " " + found_codec
            else:
                return " " + found_codec
        elif quality == 2:
            return rip_type
        else:
            return ""

    @staticmethod
    def statusFromName(name, anime=False):
        """
        Get a status object from filename

        :param name: Filename to check
        :param anime: boolean to enable anime parsing
        :return: Composite status/quality object
        """
        return Quality.compositeStatus(DOWNLOADED, Quality.nameQuality(name, anime))

    DOWNLOADED = None
    SNATCHED = None
    SNATCHED_PROPER = None
    FAILED = None
    SNATCHED_BEST = None
    ARCHIVED = None


Quality.DOWNLOADED = [Quality.compositeStatus(DOWNLOADED, x) for x in Quality.qualityStrings if x is not None]
Quality.SNATCHED = [Quality.compositeStatus(SNATCHED, x) for x in Quality.qualityStrings if x is not None]
Quality.SNATCHED_BEST = [Quality.compositeStatus(SNATCHED_BEST, x) for x in Quality.qualityStrings if x is not None]
Quality.SNATCHED_PROPER = [Quality.compositeStatus(SNATCHED_PROPER, x) for x in Quality.qualityStrings if x is not None]
Quality.FAILED = [Quality.compositeStatus(FAILED, x) for x in Quality.qualityStrings if x is not None]
Quality.ARCHIVED = [Quality.compositeStatus(ARCHIVED, x) for x in Quality.qualityStrings if x is not None]

Quality.DOWNLOADED.sort()
Quality.SNATCHED.sort()
Quality.SNATCHED_BEST.sort()
Quality.SNATCHED_PROPER.sort()
Quality.FAILED.sort()
Quality.ARCHIVED.sort()

HD720p = Quality.combineQualities([Quality.HDTV, Quality.HDWEBDL, Quality.HDBLURAY], [])
HD1080p = Quality.combineQualities([Quality.FULLHDTV, Quality.FULLHDWEBDL, Quality.FULLHDBLURAY], [])
UHD_4K = Quality.combineQualities([Quality.UHD_4K_TV, Quality.UHD_4K_WEBDL, Quality.UHD_4K_BLURAY], [])
UHD_8K = Quality.combineQualities([Quality.UHD_8K_TV, Quality.UHD_8K_WEBDL, Quality.UHD_8K_BLURAY], [])

SD = Quality.combineQualities([Quality.SDTV, Quality.SDDVD], [])
HD = Quality.combineQualities([HD720p, HD1080p], [])
UHD = Quality.combineQualities([UHD_4K, UHD_8K], [])
ANY = Quality.combineQualities([SD, HD, UHD], [])

# legacy template, cant remove due to reference in mainDB upgrade?
BEST = Quality.combineQualities([Quality.SDTV, Quality.HDTV, Quality.HDWEBDL], [Quality.HDTV])

qualityPresets = (
    ANY,
    SD,
    HD,
    HD720p,
    HD1080p,
    UHD,
    UHD_4K,
    UHD_8K,
)

qualityPresetStrings = NumDict(
    {
        SD: "SD",
        HD: "HD",
        HD720p: "HD720p",
        HD1080p: "HD1080p",
        UHD: "UHD",
        UHD_4K: "UHD-4K",
        UHD_8K: "UHD-8K",
        ANY: "Any",
    }
)


class StatusStrings(NumDict):
    """
    Dictionary containing strings for status codes
    """

    # todo: Make views return Qualities too

    qualities = Quality.DOWNLOADED + Quality.SNATCHED + Quality.SNATCHED_PROPER + Quality.SNATCHED_BEST + Quality.ARCHIVED + Quality.FAILED

    def __missing__(self, key):
        """
        If the key is not found try to determine a status from Quality

        :param key: A numeric key or None
        :raise KeyError: if the key is invalid and can't be determined from Quality
        """
        key = self.numeric(key)  # try to convert the key to a number which will raise KeyError if it can't
        if key in self.qualities:  # the key wasn't found locally so check in qualities
            status, quality = Quality.splitCompositeStatus(key)
            return self[status] if not quality else self[status] + " (" + Quality.qualityStrings[quality] + ")"
        else:  # the key wasn't found in qualities either
            raise KeyError(key)  # ... so the key is invalid

    def __contains__(self, key):
        try:
            key = self.numeric(key)
            return key in self.data or key in self.qualities
        except KeyError:
            return False


# Assign strings to statuses
statusStrings_bare = StatusStrings(
    {
        UNKNOWN: "Unknown",
        UNAIRED: "Unaired",
        SNATCHED: "Snatched",
        DOWNLOADED: "Downloaded",
        SKIPPED: "Skipped",
        SNATCHED_PROPER: "Snatched (Proper)",
        WANTED: "Wanted",
        ARCHIVED: "Archived",
        IGNORED: "Ignored",
        SUBTITLED: "Subtitled",
        FAILED: "Failed",
        SNATCHED_BEST: "Snatched (Best)",
    }
)

statusStrings = StatusStrings({key: _(value) for key, value in statusStrings_bare.items()})

cssStatusStrings = NumDict(
    {
        DOWNLOADED: "downloaded",
        SNATCHED: "snatched",
        SNATCHED_PROPER: "snatched-proper",
        FAILED: "failed",
        SNATCHED_BEST: "snatched-best",
        ARCHIVED: "archived",
        WANTED: "wanted",
    }
)


class Overview(object):
    UNAIRED = UNAIRED  # 1
    SNATCHED = SNATCHED  # 2
    WANTED = WANTED  # 3
    GOOD = DOWNLOADED  # 4
    SKIPPED = SKIPPED  # 5
    SNATCHED_PROPER = SNATCHED_PROPER  # 9
    SNATCHED_BEST = SNATCHED_BEST  # 12

    # Should suffice!
    QUAL = 50

    overviewStrings = NumDict(
        {
            SKIPPED: "skipped",
            WANTED: "wanted",
            QUAL: "qual",
            GOOD: "good",
            UNAIRED: "unaired",
            SNATCHED: "snatched",
            # we can give these a different class later, otherwise
            # breaks checkboxes in displayShow for showing different statuses
            SNATCHED_BEST: "snatched",
            SNATCHED_PROPER: "snatched",
        }
    )


countryList = {"Australia": "AU", "Canada": "CA", "USA": "US"}



class AniDBError(Exception):
    pass


class AniDBIncorrectParameterError(AniDBError):
    pass


class AniDBCommandTimeoutError(AniDBError):
    pass


class AniDBMustAuthError(AniDBError):
    pass


class AniDBPacketCorruptedError(AniDBError):
    pass


class AniDBBannedError(AniDBError):
    pass


class AniDBInternalError(AniDBError):
    pass



import re
import traceback

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("GimmePeers")

        self.urls = {
            "base_url": "https://www.gimmepeers.com",
            "login": "https://www.gimmepeers.com/takelogin.php",
            "detail": "https://www.gimmepeers.com/details.php?id=%s",
            "search": "https://www.gimmepeers.com/browse.php",
            "download": "https://gimmepeers.com/%s",
        }

        self.url = self.urls["base_url"]

        self.username = None
        self.password = None
        self.minseed = 0
        self.minleech = 0

        self.cache = tvcache.TVCache(self)

        self.search_params = {
            # c20=1&c21=1&c25=1&c24=1&c23=1&c22=1&c1=1
            "c20": 1,
            "c21": 1,
            "c25": 1,
            "c24": 1,
            "c23": 1,
            "c22": 1,
            "c1": 1,
        }

    def _check_auth(self):
        if not self.username or not self.password:
            logger.warning("Invalid username or password. Check your settings")

        return True

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {"username": self.username, "password": self.password, "ssl": "yes"}

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("Username or password incorrect!", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    logger.debug("Search string: {0}".format(search_string))

                self.search_params["search"] = search_string

                data = self.get_url(self.urls["search"], params=self.search_params, returns="text")
                if not data:
                    continue

                # Checks if cookie has timed-out causing search to redirect to login page.
                # If text matches on loginpage we login and generate a new cookie and load the search data again.
                if re.search("Still need help logging in?", data):
                    logger.debug("Login has timed out. Need to generate new cookie for GimmePeers and search again.")
                    self.session.cookies.clear()
                    self.login()

                    data = self.get_url(self.urls["search"], params=self.search_params, returns="text")
                    if not data:
                        continue

                try:
                    with BS4Parser(data, "html.parser") as html:
                        torrent_table = html.find("table", class_="browsetable")
                        torrent_rows = torrent_table("tr") if torrent_table else []

                        # Continue only if one Release is found
                        if len(torrent_rows) < 2:
                            logger.debug("Data returned from provider does not contain any torrents")
                            continue

                        for result in torrent_rows[1:]:
                            cells = result("td")

                            link = cells[1].find("a")
                            download_url = self.urls["download"] % cells[2].find("a")["href"]

                            try:
                                title = link.getText()
                                seeders = int(cells[10].getText().replace(",", ""))
                                leechers = int(cells[11].getText().replace(",", ""))
                                torrent_size = cells[8].getText()
                                size = convert_size(torrent_size) or -1
                            except (AttributeError, TypeError):
                                continue

                            if not all([title, download_url]):
                                continue

                                # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            if seeders >= 32768 or leechers >= 32768:
                                continue

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)

                except Exception:
                    logger.warning("Failed parsing provider. Traceback: {0}".format(traceback.format_exc()))

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



class Notifier(object):
    def notify_snatch(self, ep_name):
        raise NotImplementedError

    def notify_download(self, ep_name):
        raise NotImplementedError

    def notify_subtitle_download(self, ep_name, lang):
        raise NotImplementedError

    def notify_git_update(self, new_version):
        raise NotImplementedError

    def notify_login(self, ipaddress=""):
        raise NotImplementedError

    @staticmethod
    def test_notify(username):
        raise NotImplementedError



from sickchill.oldbeard.clients.generic import GenericClient


class Client(GenericClient):
    def __init__(self, host=None, username=None, password=None):

        super().__init__("mlnet", host, username, password)

        self.url = self.host
        # self.session.auth = HTTPDigestAuth(self.username, self.password);

    def _get_auth(self):

        try:
            self.response = self.session.get(self.host, verify=False)
            self.auth = self.response.content
        except Exception:
            return None

        return (None, self.auth)[self.response.status_code != 404]

    def _add_torrent_uri(self, result):

        self.url = self.host + "submit"
        params = {"q": "dllink " + result.url}
        return self._request(method="get", params=params)

    def _add_torrent_file(self, result):

        self.url = self.host + "submit"
        params = {"q": "dllink " + result.url}
        return self._request(method="get", params=params)



import re
from base64 import b16encode, b32decode
from datetime import datetime
from itertools import chain
from os.path import join
from random import shuffle

from requests.utils import add_dict_to_cookiejar

import sickchill.oldbeard
from sickchill import logger
from sickchill.helper.common import sanitize_filename
from sickchill.oldbeard.classes import Proper, SearchResult
from sickchill.oldbeard.common import MULTI_EP_RESULT, Quality, SEASON_RESULT
from sickchill.oldbeard.db import DBConnection
from sickchill.oldbeard.helpers import download_file, getURL, make_session, remove_file_failed
from sickchill.oldbeard.name_parser.parser import InvalidNameException, InvalidShowException, NameParser
from sickchill.oldbeard.show_name_helpers import allPossibleShowNames
from sickchill.oldbeard.tvcache import TVCache


class GenericProvider(object):
    NZB = "nzb"
    NZBDATA = "nzbdata"
    TORRENT = "torrent"

    PROVIDER_BROKEN = 0
    PROVIDER_DAILY = 1
    PROVIDER_BACKLOG = 2
    PROVIDER_OK = 3

    ProviderStatus = {
        PROVIDER_BROKEN: _("Not working"),
        PROVIDER_DAILY: _("Daily/RSS only"),
        PROVIDER_BACKLOG: _("Backlog/Manual Search only"),
        PROVIDER_OK: _("Daily/RSS and Backlog/Manual Searches working"),
    }

    def __init__(self, name):
        self.name = name

        self.anime_only = False
        self.bt_cache_urls = [
            # 'http://torcache.net/torrent/{torrent_hash}.torrent',
            # 'http://torrentproject.se/torrent/{torrent_hash}.torrent',
            "http://thetorrent.org/torrent/{torrent_hash}.torrent",
            # 'http://btdig.com/torrent/{torrent_hash}.torrent',
            ("https://torrage.info/download?h={torrent_hash}", "https://torrage.info/torrent.php?h={torrent_hash}"),
            "https://itorrents.org/torrent/{torrent_hash}.torrent?title={torrent_name}",
        ]
        self.cache = TVCache(self)
        self.enable_backlog = False
        self.enable_daily = False
        self.enabled = False
        self.headers = dict()
        self.proper_strings = ["PROPER|REPACK|REAL"]
        self.provider_type = None
        self.public = False
        self.search_fallback = False
        self.search_mode = None
        self.session = make_session()
        self.show = None
        self.supports_absolute_numbering = False
        self.supports_backlog = True
        self.supports_movies = False

        self.url = ""
        self.urls = {}

        # Use and configure the attribute enable_cookies to show or hide the cookies input field per provider
        self.enable_cookies = False
        self.cookies = ""
        self.rss_cookies = ""

        self.ability_status = self.PROVIDER_OK

        shuffle(self.bt_cache_urls)

    def download_result(self, result):
        if not self.login():
            return False

        urls, filename = self._make_url(result)

        for url in urls:
            if "NO_DOWNLOAD_NAME" in url:
                continue

            if isinstance(url, tuple):
                referer = url[1]
                url = url[0]
            else:
                referer = "/".join(url.split("/")[:3]) + "/"

            if url.startswith("http"):
                self.headers.update({"Referer": referer})

            logger.info("Downloading a result from {0} at {1}".format(self.name, url))

            downloaded_filename = download_file(
                url, filename, session=self.session, headers=self.headers, hooks={"response": self.get_url_hook}, return_filename=True
            )
            if downloaded_filename:
                if self._verify_download(downloaded_filename):
                    logger.info("Saved result to {0}".format(downloaded_filename))
                    return True

                logger.warning("Could not download {0}".format(url))
                remove_file_failed(downloaded_filename)

        if urls:
            logger.warning("Failed to download any results")

        return False

    def find_propers(self, search_date=None):
        results = self.cache.list_propers(search_date)

        return [Proper(x["name"], x["url"], datetime.fromtimestamp(x["time"]), self.show) for x in results]

    def find_search_results(self, show, episodes, search_mode, manual_search=False, download_current_quality=False):
        self._check_auth()
        self.show = show

        results = {}
        items_list = []
        searched_scene_season = None

        for episode in episodes:
            cache_result = self.cache.search_cache(episode, manual_search=manual_search, down_cur_quality=download_current_quality)
            if cache_result:
                if episode.episode not in results:
                    results[episode.episode] = cache_result
                else:
                    results[episode.episode].extend(cache_result)

                continue

            if len(episodes) > 1 and search_mode == "sponly" and searched_scene_season == episode.scene_season:
                continue

            search_strings = []
            searched_scene_season = episode.scene_season

            if len(episodes) > 1 and search_mode == "sponly":
                search_strings = self.get_season_search_strings(episode)
            elif search_mode == "eponly":
                search_strings = self.get_episode_search_strings(episode)

            for search_string in search_strings:
                items_list += self.search(search_string, ep_obj=episode)

        if len(results) == len(episodes):
            return results

        if items_list:
            items = {}
            unknown_items = []

            for item in items_list:
                quality = self.get_quality(item, anime=show.is_anime)

                if quality == Quality.UNKNOWN:
                    unknown_items.append(item)
                elif quality == Quality.NONE:
                    pass  # Skipping an HEVC when HEVC is not allowed by settings
                else:
                    if quality not in items:
                        items[quality] = []
                    items[quality].append(item)

            items_list = list(chain(*[v for (k_, v) in sorted(items.items(), reverse=True)]))
            items_list += unknown_items

        cl = []

        for item in items_list:
            title, url = self._get_title_and_url(item)
            seeders, leechers = self._get_seeders_and_leechers(item)
            size = self._get_size(item)

            try:
                parse_result = NameParser(parse_method=("normal", "anime")[show.is_anime]).parse(title)
            except (InvalidNameException, InvalidShowException) as error:
                logger.debug("{0}".format(error))
                continue

            show_object = parse_result.show
            quality = parse_result.quality
            release_group = parse_result.release_group
            version = parse_result.version
            skip_release = False
            actual_episodes = []
            actual_season = -1

            if not (show_object.air_by_date or show_object.sports):
                if search_mode == "sponly":
                    if parse_result.episode_numbers:
                        logger.debug("This is supposed to be a season pack search but the result {0} is not a valid season pack, skipping it".format(title))
                        skip_release = True
                    elif not [ep for ep in episodes if parse_result.season_number == (ep.season, ep.scene_season)[ep.show.is_scene]]:
                        logger.info("This season result {0} is for a season we are not searching for, skipping it".format(title), logger.DEBUG)
                        skip_release = True

                else:
                    if not all(
                        [
                            parse_result.season_number is not None,
                            parse_result.episode_numbers,
                            [
                                ep
                                for ep in episodes
                                if (ep.season, ep.scene_season)[ep.show.is_scene] == (parse_result.season_number, parse_result.scene_season)[ep.show.is_scene]
                                and (ep.episode, ep.scene_episode)[ep.show.is_scene] in parse_result.episode_numbers
                            ],
                        ]
                    ) and not all(
                        [
                            # fallback for anime on absolute numbering
                            parse_result.is_anime,
                            parse_result.ab_episode_numbers is not None,
                            [ep for ep in episodes if ep.show.is_anime and ep.absolute_number in parse_result.ab_episode_numbers],
                        ]
                    ):

                        logger.info("The result {0} doesn't seem to match an episode that we are currently trying to snatch, skipping it".format(title))
                        skip_release = True

                if not skip_release:
                    actual_season = parse_result.season_number
                    actual_episodes = parse_result.episode_numbers
            else:
                same_day_special = False

                if not parse_result.is_air_by_date:
                    logger.debug("This is supposed to be a date search but the result {0} didn't parse as one, skipping it".format(title))
                    skip_release = True
                else:
                    air_date = parse_result.air_date.toordinal()
                    db = DBConnection()
                    sql_results = db.select("SELECT season, episode FROM tv_episodes WHERE showid = ? AND airdate = ?", [show_object.indexerid, air_date])

                    if len(sql_results) == 2:
                        if int(sql_results[0]["season"]) == 0 and int(sql_results[1]["season"]) != 0:
                            actual_season = int(sql_results[1]["season"])
                            actual_episodes = [int(sql_results[1]["episode"])]
                            same_day_special = True
                        elif int(sql_results[1]["season"]) == 0 and int(sql_results[0]["season"]) != 0:
                            actual_season = int(sql_results[0]["season"])
                            actual_episodes = [int(sql_results[0]["episode"])]
                            same_day_special = True
                    elif len(sql_results) != 1:
                        logger.warning("Tried to look up the date for the episode {0} but the database didn't give proper results, skipping it".format(title))
                        skip_release = True

                    if not skip_release and not same_day_special:
                        actual_season = int(sql_results[0]["season"])
                        actual_episodes = [int(sql_results[0]["episode"])]

            logger.debug("Adding item from search to cache: {0}".format(title))

            ci = self.cache._add_cache_entry(title, url, size, seeders, leechers, parse_result=parse_result)

            if ci is not None:
                cl.append(ci)

            if skip_release:
                continue

            for episode_number in actual_episodes:
                if not show_object.wantEpisode(actual_season, episode_number, quality, manual_search, download_current_quality):
                    skip_release = True
                    break

            if skip_release:
                logger.debug(_("Ignoring result ") + f"{title}.")
                continue

            logger.debug(_("Found result {title} at {url}".format(title=title, url=url)))

            episode_object = []
            for current_episode in actual_episodes:
                episode_object.append(show_object.getEpisode(actual_season, current_episode))

            result = self.get_result(episode_object)
            result.show = show_object
            result.url = url
            result.name = title
            result.quality = quality
            result.release_group = release_group
            result.version = version
            result.content = None
            result.size = self._get_size(item)

            if len(episode_object) == 1:
                episode_number = episode_object[0].episode
                logger.debug("Single episode result.")
            elif len(episode_object) > 1:
                episode_number = MULTI_EP_RESULT
                logger.debug("Separating multi-episode result to check for later - result contains episodes: {0}".format(parse_result.episode_numbers))
            elif len(episode_object) == 0:
                episode_number = SEASON_RESULT
                logger.debug("Separating full season result to check for later")
            else:
                episode_number = None

            if episode_number not in results:
                results[episode_number] = [result]
            else:
                results[episode_number].append(result)

        if cl:

            # Access to a protected member of a client class
            cache_db = self.cache._get_db()
            cache_db.mass_upsert("results", cl)

        return results

    def get_id(self, suffix=""):
        return GenericProvider.make_id(self.name) + str(suffix)

    def get_quality(self, item, anime=False):
        (title, url_) = self._get_title_and_url(item)
        quality = Quality.scene_quality(title, anime)

        return quality

    def get_result(self, episodes):
        result = self._get_result(episodes)
        result.provider = self

        return result

    @staticmethod
    def get_url_hook(response, **kwargs_):
        if response:
            logger.debug("{0} URL: {1} [Status: {2}]".format(response.request.method, response.request.url, response.status_code))

            if response.request.method == "POST":
                logger.debug("With post data: {0}".format(response.request.body))

    def get_url(self, url, post_data=None, params=None, timeout=30, **kwargs):
        kwargs["hooks"] = {"response": self.get_url_hook}
        return getURL(url, post_data, params, self.headers, timeout, self.session, **kwargs)

    def image_name(self):
        return self.get_id() + ".png"

    @property
    def is_active(self):
        return False

    @property
    def is_enabled(self):
        return bool(self.enabled)

    @property
    def daily_enabled(self):
        return int(self.enable_daily)

    @property
    def backlog_enabled(self):
        return int(self.enable_backlog)

    @property
    def search_fallback_enabled(self):
        return int(self.search_fallback)

    @property
    def can_daily(self):
        return self.ability_status & self.PROVIDER_DAILY != 0

    @property
    def can_backlog(self):
        return self.ability_status & self.PROVIDER_BACKLOG != 0 and self.supports_backlog

    def status(self):
        return self.ProviderStatus.get(self.ability_status)

    @staticmethod
    def make_id(name):
        if not name:
            return ""

        return re.sub(r"[^\w\d_]", "_", str(name).strip().lower())

    def search_rss(self, episodes):
        return self.cache.find_needed_episodes(episodes)

    def seed_ratio(self):
        return ""

    def _check_auth(self):
        return True

    def login(self):
        return True

    def search(self, search_params, age=0, ep_obj=None):
        return []

    def _get_result(self, episodes):
        return SearchResult(episodes)

    def get_episode_search_strings(self, episode, add_string=""):
        if not episode:
            return []

        search_string = {"Episode": []}

        for show_name in allPossibleShowNames(episode.show, season=episode.scene_season):
            episode_string = show_name + " "
            episode_string_fallback = None

            if episode.show.air_by_date:
                episode_string += str(episode.airdate).replace("-", " ")
            elif episode.show.sports:
                episode_string += str(episode.airdate).replace("-", " ")
                episode_string += ("|", " ")[len(self.proper_strings) > 1]
                episode_string += episode.airdate.strftime("%b")
            elif episode.show.anime:
                episode_string_fallback = episode_string + "{0:02d}".format(int(episode.scene_absolute_number))
                episode_string += "{0:03d}".format(int(episode.scene_absolute_number))
            else:
                episode_string += sickchill.oldbeard.config.naming_ep_type[2] % {
                    "seasonnumber": episode.scene_season,
                    "episodenumber": episode.scene_episode,
                }

            if add_string:
                episode_string += " " + add_string
                if episode_string_fallback:
                    episode_string_fallback += " " + add_string

            search_string["Episode"].append(episode_string.strip())
            if episode_string_fallback:
                search_string["Episode"].append(episode_string_fallback.strip())

        return [search_string]

    def get_season_search_strings(self, episode):
        search_string = {"Season": []}

        for show_name in allPossibleShowNames(episode.show, season=episode.scene_season):
            season_string = show_name + " "

            if episode.show.air_by_date or episode.show.sports:
                season_string += str(episode.airdate).split("-")[0]
            elif episode.show.anime:
                # use string below if you really want to search on season with number
                # season_string += 'Season ' + '{0:d}'.format(int(episode.scene_season))
                season_string += "Season"  # ignore season number to get all seasons in all formats
            else:
                season_string += "S{0:02d}".format(int(episode.scene_season))

            search_string["Season"].append(season_string.strip())

        return [search_string]

    def _get_size(self, item):
        try:
            return item.get("size", -1)
        except AttributeError:
            return -1

    def _get_seeders_and_leechers(self, item):
        try:
            return item.get("seeders", -1), item.get("leechers", -1)
        except AttributeError:
            return -1, -1

    def _get_storage_dir(self):
        return ""

    def _get_title_and_url(self, item):
        if not item:
            return "", ""

        title = item.get("title", "")
        url = item.get("link", "")

        if title:
            title = "" + title.replace(" ", ".")
        else:
            title = ""

        if url:
            url = url.replace("&amp;", "&").replace("%26tr%3D", "&tr=")
        else:
            url = ""

        return title, url

    @staticmethod
    def hash_from_magnet(magnet):
        try:
            torrent_hash = re.findall(r"urn:btih:([\w]{32,40})", magnet)[0].upper()
            if len(torrent_hash) == 32:
                torrent_hash = b16encode(b32decode(torrent_hash)).upper()
            return torrent_hash
        except Exception:
            logger.exception("Unable to extract torrent hash or name from magnet: {0}".format(magnet))
            return ""

    def _make_url(self, result):
        if not result:
            return "", ""

        filename = ""

        result.url = result.url.replace("http://itorrents.org", "https://itorrents.org")

        urls = [result.url]
        if result.url.startswith("magnet"):
            torrent_hash = self.hash_from_magnet(result.url)
            if not torrent_hash:
                return urls, filename

            try:
                torrent_name = re.findall("dn=([^&]+)", result.url)[0]
            except Exception:
                torrent_name = "NO_DOWNLOAD_NAME"

            urls = []
            for cache_url in self.bt_cache_urls:
                if isinstance(cache_url, tuple):
                    urls.append(
                        (
                            cache_url[0].format(torrent_hash=torrent_hash, torrent_name=torrent_name),
                            cache_url[1].format(torrent_hash=torrent_hash, torrent_name=torrent_name),
                        )
                    )
                else:
                    urls.append(cache_url.format(torrent_hash=torrent_hash, torrent_name=torrent_name))

        if "torrage.info/torrent.php" in result.url:
            torrent_hash = result.url.split("=")[1]
            urls = [
                (
                    "https://t.torrage.info/download?h={torrent_hash}".format(torrent_hash=torrent_hash),
                    "https://torrage.info/torrent.php?h={torrent_hash}".format(torrent_hash=torrent_hash),
                )
            ]

        filename = join(self._get_storage_dir(), sanitize_filename(result.name) + "." + self.provider_type)

        return urls, filename

    def _verify_download(self, filename):
        return True

    def add_cookies_from_ui(self):
        """
        Adds the cookies configured from UI to the providers requests session
        :return: A tuple with the the (success result, and a descriptive message in str)
        """

        # This is the generic attribute used to manually add cookies for provider authentication
        if self.enable_cookies and self.cookies:
            cookie_validator = re.compile(r"^(\w+=[\w%-]+)(;\w+=[\w%-]+)*$")
            if not cookie_validator.match(self.cookies):
                return False, "Cookie is not correctly formatted: {0}".format(self.cookies)

            new_cookies = {}
            for cookie in self.cookies.split(";"):
                key, value = cookie.rsplit("=", 1)
                new_cookies[key] = value

            add_dict_to_cookiejar(self.session.cookies, new_cookies)
            return True, "torrent cookie"

        return False, "No Cookies added from ui for provider: {0}".format(self.name)



import datetime
import logging

import guessit
from slugify import slugify
from sqlalchemy import Boolean, Column, Date, DateTime, ForeignKey, Integer, Interval, JSON, SmallInteger, String
from sqlalchemy.event import listen
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker

logger = logging.getLogger("sickchill.movie")

Base = declarative_base()
Session = sessionmaker()


class Movie(Base):
    __tablename__ = "movie"
    pk = Column(Integer, primary_key=True)
    name = Column(String)
    date = Column(Date)
    year = Column(SmallInteger)
    status = Column(Integer)
    paused = Column(Boolean, default=False)
    location = Column(String)
    start = Column(Interval, default=datetime.timedelta(days=-7))
    interval = Column(Interval, default=datetime.timedelta(days=1))
    added = Column(DateTime, default=datetime.datetime.now)
    updated = Column(DateTime, onupdate=datetime.datetime.now)
    completed = Column(DateTime)
    searched = Column(DateTime)
    slug = Column(String)

    language = Column(String)

    result_pk = Column(Integer, ForeignKey("result.pk"))
    results: list = relationship("Result", back_populates="movie")

    images: list = relationship("Images", backref="movie")
    indexer_data: list = relationship("IndexerData", backref="movie")

    def __init__(self, name: str, year: int):
        self.name = name
        self.year = year

    @property
    def poster(self):
        return ""

    def __get_named_indexer_data(self, name):
        if self.indexer_data:
            for data in self.indexer_data:
                if data.site == name:
                    return data

    @property
    def imdb_data(self):
        data = self.__get_named_indexer_data("imdb")
        if data:
            return data.data
        return dict()

    @property
    def imdb_id(self):
        data = self.__get_named_indexer_data("imdb")
        if data:
            return data.pk
        return ""

    @property
    def tmdb_id(self):
        data = self.__get_named_indexer_data("tmdb")
        if data:
            return data.pk
        return ""

    @property
    def imdb_genres(self):
        data = self.__get_named_indexer_data("imdb")
        if data:
            return data.genres
        return []

    def __get_indexer_values(self, name, keys: list):
        try:
            data = getattr(self, f"{name}_data")
            for key in keys:
                data = data[key]
            return data
        except AttributeError:
            logger.debug(f"We do not have data for {name}")
        except (IndexError, KeyError):
            logger.debug(f"KeyError: {name}{''.join([f'[{k}]' for k in keys])}")

    @property
    def runtime(self):
        return self.__get_indexer_values("imdb", ["base", "runningTimeInMinutes"])

    @property
    def imdb_votes(self):
        return self.__get_indexer_values("imdb", ["ratings", "ratingCount"])

    @property
    def imdb_rating(self):
        return self.__get_indexer_values("imdb", ["ratings", "rating"])

    @property
    def imdb_outline(self):
        return self.__get_indexer_values("imdb", ["plot", "outline", "text"])

    @property
    def imdb_summary(self):
        return self.__get_indexer_values("imdb", ["plot", "summaries", 0, "text"])

    @staticmethod
    def slugify(target, value, old_value, initiator):
        if value and (not target.slug or value != old_value):
            target.slug = slugify(value)

    def search_strings(self):
        return {"Movie": [f"{self.name} {self.year}"]}

    def __repr__(self):
        return f"{self.name}"


listen(Movie.name, "set", Movie.slugify, retval=False)


class Result(Base):
    __tablename__ = "result"
    pk = Column(Integer, primary_key=True)
    name = Column(String)
    title = Column(String)
    url = Column(String)
    size = Column(Integer)
    year = Column(SmallInteger)
    provider = Column(String)
    seeders = Column(Integer)
    leechers = Column(Integer)
    info_hash = Column(String)
    group = Column(String)
    type = Column(String)
    guess = Column(JSON)
    found = Column(DateTime, default=datetime.datetime.now)
    updated = Column(DateTime, onupdate=datetime.datetime.now)

    movie_pk = Column(Integer, ForeignKey("movie.pk"))
    movie = relationship("Movie", back_populates="result")

    session = Session()

    def __init__(self, result: dict, movie: Movie, provider):
        name = result["title"]
        guess = guessit.guessit(name)
        if not (guess and guess["type"] == "movie"):
            logging.debug(f"This is an episode, not a movie: {name}")
            return

        if not self.session.query(Movie).filter(Movie.name.like(f"{guess['title']}%")).count():
            logging.debug(f"This result does not match any of our movies")
            return

        self.info_hash = result["hash"]
        self.url = result["link"]
        self.name = name
        self.title = guess["title"]
        self.group = guess["release_group"]
        self.seeders = result["seeders"]
        self.leechers = result["leechers"]
        self.size = result["size"]
        self.year = guess["year"] or movie.year
        self.type = provider.provider_type

        self.provider = provider.get_id()

        self.guess = guess

        self.movie = movie

    def __repr__(self):
        return f"{self.name}"


class Images(Base):
    __tablename__ = "images"

    url = Column(String, primary_key=True)
    path = Column(String)
    site = Column(String)
    style = Column(Integer)

    movie_pk = Column(Integer, ForeignKey("movie.pk"))

    def __init__(self, site: str, movie_pk: int, url: str, path: str, style: int):
        self.url = url
        self.path = path
        self.site = site
        self.style = style
        self.movie_pk = movie_pk


class IndexerData(Base):
    __tablename__ = "indexer_data"
    pk = Column(String, primary_key=True)
    site = Column(String)
    data = Column(JSON)

    movie_pk = Column(Integer, ForeignKey("movie.pk"))

    genres: list = relationship("Genres", backref="indexer_data")

    def __repr__(self):
        return f"[{self.__tablename__.replace('_', ' ').title()}] {self.site}: {self.pk} - {self.movie.name}"


class Genres(Base):
    __tablename__ = "genres"
    pk = Column(String, primary_key=True)
    indexer_data_pk = Column(Integer, ForeignKey("indexer_data.pk"))



import traceback

from rtorrent import RTorrent

from sickchill import logger, settings
from sickchill.oldbeard.clients.generic import GenericClient


class Client(GenericClient):
    def __init__(self, host=None, username=None, password=None):
        super().__init__("rTorrent", host, username, password)

    def _get_auth(self):

        if self.auth is not None:
            return self.auth

        if not self.host:
            return

        self.host = self.host.rstrip("/")

        tp_kwargs = {}
        if settings.TORRENT_AUTH_TYPE and settings.TORRENT_AUTH_TYPE.lower() != "none":
            tp_kwargs["authtype"] = settings.TORRENT_AUTH_TYPE

        if not settings.TORRENT_VERIFY_CERT:
            tp_kwargs["check_ssl_cert"] = False

        if self.username and self.password:
            self.auth = RTorrent(self.host, self.username, self.password, True, tp_kwargs=tp_kwargs or None)
        else:
            self.auth = RTorrent(self.host, None, None, True, tp_kwargs=tp_kwargs or None)

        return self.auth

    def _add_torrent_uri(self, result):

        if not (self.auth and result):
            return False

        try:
            # Send torrent magnet with params to rTorrent and optionally start download
            torrent = self.auth.load_magnet(result.url, result.hash, start=not settings.TORRENT_PAUSED, params=self._get_params(result))

            if not torrent:
                return False

            return True

        except Exception as error:
            logger.warning(_("Error while sending torrent: {error}".format(error=error)))
            return False

    def _add_torrent_file(self, result):

        if not (self.auth and result):
            return False

        try:
            # Send torrent file with params to rTorrent and optionally start download
            torrent = self.auth.load_torrent(result.content, start=not settings.TORRENT_PAUSED, params=self._get_params(result))

            if not torrent:
                return False

            return True

        except Exception as error:
            logger.info(traceback.format_exc())
            logger.warning(_("Error while sending torrent: {error}".format(error=error)))
            return False

    def testAuthentication(self):
        try:
            self._get_auth()

            if self.auth is not None:
                return True, _("Success: Connected and Authenticated")
            else:
                return False, _("Error: Unable to get {self.name} Authentication, check your config!")
        except Exception as error:
            return False, _("Error: Unable to connect to {name}: {error}".format(name=self.name, error=error))

    @staticmethod
    def _get_params(result):
        params = []

        # Set label
        label = settings.TORRENT_LABEL
        if result.show.is_anime:
            label = settings.TORRENT_LABEL_ANIME
        if label:
            params.append(f"d.custom1.set={label}")

        # Set download folder
        if settings.TORRENT_PATH:
            params.append(f"d.directory.set={settings.TORRENT_PATH}")

        return params



from urllib.error import URLError
from urllib.parse import unquote_plus

from beekeeper.exceptions import RequestTimeout
from kodipydent import Kodi

from sickchill import logger, settings
from sickchill.helper import try_int
from sickchill.oldbeard import common


class Notifier(object):
    def __init__(self):
        self._connections = list()

    def setup(self, hosts=None, username=None, password=None):
        self._connections = []
        for host in (x.strip() for x in (hosts or settings.KODI_HOST or "").split(",") if x.strip()):
            try:
                if ":" in host:
                    bare_host, port = host.split(":")
                    kodi = Kodi(bare_host, username or settings.KODI_USERNAME, password or settings.KODI_PASSWORD, port=try_int(port, 8080))
                else:
                    kodi = Kodi(host, username or settings.KODI_USERNAME, password or settings.KODI_PASSWORD)
                kodi.host = kodi.variables()["hostname"]["value"]
                kodi.name = kodi.Settings.GetSettingValue(setting="services.devicename")["result"]["value"]
                if kodi not in self._connections:
                    self._connections.append(kodi)
            except (URLError, RequestTimeout, KeyError, IndexError):
                pass
                # logger.info('Unable to connect to Kodi host as {0}, make sure the username and password is correct and that the http control is enabled'.format(host))

    @property
    def connections(self):
        if not self._connections:
            self.setup()
        return self._connections

    @connections.setter
    def connections(self, value):
        self._connections = value

    @staticmethod
    def success(result):
        try:
            return result["result"] == "OK"
        except (AttributeError, KeyError, ValueError):
            print(result["error"])

        return False

    def _notify_kodi(self, message, title="SickChill", hosts=None, username=None, password=None, force=False, dest_app="KODI"):
        """Internal wrapper for the notify_snatch and notify_download functions

        Detects JSON-RPC version then branches the logic for either the JSON-RPC or legacy HTTP API methods.

        Args:
            message: Message body of the notice to send
            title: Title of the notice to send
            hosts: KODI webserver host:port
            username: KODI webserver username
            password: KODI webserver password
            force: Used for the Test method to override config saftey checks

        Returns:
            Returns a list results in the format of host:ip:result
            The result will either be 'OK' or False, this is used to be parsed by the calling function.

        """

        self.setup(hosts=hosts, username=username, password=password)
        results = dict()
        for connection in self.connections:
            logger.debug("Sending {0} notification to '{1}' - {2}".format(dest_app, connection.host, message))
            response = connection.GUI.ShowNotification(title=title, message=message, image=settings.LOGO_URL)
            if response and response.get("result"):
                results[connection.host] = self.success(response)
            else:
                if settings.KODI_ALWAYS_ON or force:
                    logger.warning("Failed to send notification to {0} for '{1}', check configuration and try again.".format(dest_app, connection.host))
                results[connection.host] = False

        for host in [x.strip() for x in (hosts or settings.KODI_HOST or "").split(",") if x.strip()]:
            base_host = host.split(":")[0]
            if base_host not in results and host not in results:
                results[host] = False

        return results

    def update_library(self, show_name=None):
        if not (settings.USE_KODI and settings.KODI_UPDATE_LIBRARY):
            return False

        if not settings.KODI_HOST:
            logger.debug("No KODI hosts specified, check your settings")
            return False

        result = 0
        for connection in self.connections:
            try:
                logger.debug("Updating KODI library on host: " + connection.host)

                if show_name:
                    tvshowid = -1
                    path = ""

                    logger.debug("Updating library in KODI for show " + show_name)

                    # let's try letting kodi filter the shows
                    response = connection.VideoLibrary.GetTVShows(
                        filter={"field": "title", "operator": "is", "value": unquote_plus(show_name)}, properties=["file"]
                    )
                    if response and "result" in response and "tvshows" in response["result"]:
                        shows = response["result"]["tvshows"]
                    else:
                        # fall back to retrieving the entire show list
                        response = connection.VideoLibrary.GetTVShows(properties=["file"])
                        if response and "result" in response and "tvshows" in response["result"]:
                            shows = response["result"]["tvshows"]
                        else:
                            logger.debug("KODI: No tvshows in KODI TV show list")
                            continue

                    check = (show_name, unquote_plus(show_name))
                    for show in shows:
                        if ("label" in show and show["label"] in check) or ("title" in show and show["title"] in check):
                            tvshowid = show["tvshowid"]
                            path = show["file"]
                            break

                    del shows

                    # we didn't find the show (exact match), thus revert to just doing a full update if enabled
                    if tvshowid == -1:
                        logger.debug("Exact show name not matched in KODI TV show list")

                    if not path:
                        logger.warning("No valid path found for " + show_name + " with ID: " + str(tvshowid) + " on " + connection.host)

                    if path and tvshowid != -1:
                        logger.debug("KODI Updating " + show_name + " with ID: " + str(tvshowid) + " at " + path + " on " + connection.host)
                        response = connection.VideoLibrary.Scan(directory=path)
                        if not self.success(response):
                            logger.warning("Update of show directory failed on " + show_name + " on " + connection.host + " at " + path)
                        else:
                            if settings.KODI_UPDATE_ONLYFIRST:
                                logger.debug("Successfully updated '" + connection.host + "', stopped sending update library commands.")
                                return True

                            result += 1
                            continue

                if show_name and not settings.KODI_UPDATE_FULL:
                    continue

                logger.debug("Doing Full Library KODI update on host: " + connection.host)
                response = connection.VideoLibrary.Scan()
                if not response:
                    logger.warning("KODI Full Library update failed on: " + connection.host)
                elif self.success(response):
                    if settings.KODI_UPDATE_ONLYFIRST:
                        logger.debug("Successfully updated '" + connection.host + "', stopped sending update library commands.")
                        return True

                    result += 1
            except (URLError, RequestTimeout):
                pass

        return result > 0

    def play_episode(self, episode, connection_index=0):
        """Handles playing videos on a KODI host via HTTP JSON-RPC

        Attempts to play an episode on a KODI host.

        Args:
            episode: The episode to play
            connection_index: Index of the selected host to play the episode on

        Returns:
            Returns True or False

        """
        try:
            connection = self.connections[int(connection_index)]
        except IndexError:
            logger.warning("Incorrect KODI host passed to play an episode, aborting play")
            return False

        logger.debug("Trying to play episode on Kodi for host: " + connection.host)

        response = connection.VideoLibrary.GetTVShows(filter={"field": "title", "operator": "is", "value": episode.show.name})

        shows = []
        tvshowid = None
        if response and "result" in response and "tvshows" in response["result"]:
            shows = response["result"]["tvshows"]

        check = (episode.show.name, unquote_plus(episode.show.name))
        for show in shows:
            if ("label" in show and show["label"] in check) or ("title" in show and show["title"] in check):
                tvshowid = show["tvshowid"]

        del shows

        if tvshowid is None:
            logger.info("Could not play the item, could not find the show on Kodi")
            return

        response = connection.VideoLibrary.GetEpisodes(
            filter={"field": "title", "operator": "is", "value": episode.name}, season=episode.season, tvshowid=tvshowid, properties=["file"]
        )

        if response and "result" in response and "episodes" in response["result"]:
            episodes = response["result"]["episodes"]

            if len(episodes) > 1:
                logger.info("Could not play the item, too many files were returned as options and we could not choose")

            if episodes:
                connection.Player.Open(item={"file": episodes[0]["file"]})
            else:
                logger.info("Could not find the episode on Kodi to play")

    def notify_snatch(self, ep_name):
        if settings.KODI_NOTIFY_ONSNATCH:
            self._notify_kodi(ep_name, common.notifyStrings[common.NOTIFY_SNATCH])

    def notify_download(self, ep_name):
        if settings.KODI_NOTIFY_ONDOWNLOAD:
            self._notify_kodi(ep_name, common.notifyStrings[common.NOTIFY_DOWNLOAD])

    def notify_subtitle_download(self, ep_name, lang):
        if settings.KODI_NOTIFY_ONSUBTITLEDOWNLOAD:
            self._notify_kodi(ep_name + ": " + lang, common.notifyStrings[common.NOTIFY_SUBTITLE_DOWNLOAD])

    def notify_git_update(self, new_version="??"):
        if settings.USE_KODI:
            update_text = common.notifyStrings[common.NOTIFY_GIT_UPDATE_TEXT]
            title = common.notifyStrings[common.NOTIFY_GIT_UPDATE]
            self._notify_kodi(update_text + new_version, title)

    def notify_login(self, ipaddress=""):
        if settings.USE_KODI:
            update_text = common.notifyStrings[common.NOTIFY_LOGIN_TEXT]
            title = common.notifyStrings[common.NOTIFY_LOGIN]
            self._notify_kodi(update_text.format(ipaddress), title)

    def test_notify(self, host, username, password):
        return self._notify_kodi("Testing KODI notifications from SickChill", "Test Notification", hosts=host, username=username, password=password, force=True)



import datetime
import time
from urllib.parse import urljoin

import js2py

from sickchill import logger
from sickchill.helper.common import try_int
from sickchill.oldbeard import db, tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("ThePirateBay")

        # Credentials
        self.public = True

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0
        self.confirmed = True

        # URLs
        self.url = "https://thepiratebay.org"
        self.api = "https://apibay.org"
        self.queries = {"top": ["top100:208", "top100:205"]}
        self.script_url = "https://torrindex.net/static/main.js"

        # "https://apibay.org/precompiled/data_top100_48h_205.json"
        # "https://apibay.org/precompiled/data_top100_48h_208.json"
        #
        # "https://apibay.org/q.php?q=Arrow S08E08&cat=205,208"
        # "https://apibay.org/q.php?q=tt2193021&cats=205,208"

        self.urls = {
            "search": urljoin(self.api, "q.php"),
            "rss": [urljoin(self.api, "precompiled/data_top100_48h_208.json"), urljoin(self.api, "precompiled/data_top100_48h_205.json")],
        }

        # Cache
        self.cache = tvcache.TVCache(self, min_time=30)  # only poll ThePirateBay every 30 minutes max

    @property
    def tracker_cache(self):
        return TrackerCacheDBConnection(self)

    def get_tracker_list(self):
        try:
            script = self.get_url(self.script_url)
            context = js2py.EvalJs()
            context.execute(
                """
                escape = function(text){pyimport urllib; return urllib.quote(text)};
                unescape = function(text){pyimport urllib; return urllib.unquote(text)};
                encodeURI = function(text){pyimport urllib; return urllib.quote(text, safe='~@#$&()*!+=:;,.?/\\'')};
                decodeURI = unescape;
                encodeURIComponent = function(text){pyimport urllib; return urllib.quote(text, safe='~()*!.\\'')};
                decodeURIComponent = unescape;
                """
            )

            context.execute(script)
            return context.print_trackers()
        except Exception:
            return ""

    def make_magnet(self, name, info_hash):
        trackers = self.tracker_cache.get_trackers()
        return "magnet:?xt=urn:btih:{info_hash}&dn={name}{trackers}".format(name=name, info_hash=info_hash, trackers=trackers) + self._custom_trackers

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        search_params = {"cat": "208,205", "q": None}

        if not (self.tracker_cache.get_trackers() or self._custom_trackers):
            logger.info("Set some custom trackers in config/search on the torrents tab. Re-enable this provider after fixing this issue.")
            self.enabled = False
            return results

        for mode in search_strings:
            items = []
            logger.debug("Search Mode: {0}".format(mode))

            all_search_strings = search_strings[mode]
            if mode != "RSS" and self.show and self.show.imdb_id:
                all_search_strings = [self.show.imdb_id] + search_strings[mode]

            for search_string in all_search_strings:
                search_urls = (self.urls["search"], self.urls["rss"])[mode == "RSS"]
                if not isinstance(search_urls, list):
                    search_urls = [search_urls]

                for search_url in search_urls:
                    if mode != "RSS":
                        search_params["q"] = search_string
                        logger.debug("Search string: {}".format(search_string))

                        data = self.get_url(search_url, params=search_params, returns="json")
                    else:
                        data = self.get_url(search_url, returns="json")

                    if not (data and isinstance(data, list)):
                        logger.debug("URL did not return data")
                        continue

                    for result in data:
                        try:
                            title = result["name"]
                            if title == "No results returned":
                                logger.debug(title)
                                continue

                            info_hash = result["info_hash"]
                            if not all([title, info_hash]):
                                continue

                            seeders = result["seeders"]
                            leechers = result["leechers"]

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            # Accept Torrent only from Good People for every Episode Search
                            if self.confirmed and not result["status"] in ("trusted", "vip"):
                                if mode != "RSS":
                                    logger.debug("Found result: {0} but that doesn't seem like a trusted result so I'm ignoring it".format(title))
                                continue

                            torrent_size = try_int(result["size"])

                            item = {
                                "title": title,
                                "link": self.make_magnet(title, info_hash),
                                "size": torrent_size,
                                "seeders": seeders,
                                "leechers": leechers,
                                "hash": info_hash,
                            }
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)
                        except Exception:
                            continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results


class TrackerCacheDBConnection(db.DBConnection):
    def __init__(self, provider_instance=None, update_frequency=datetime.timedelta(hours=8)):
        super().__init__("cache.db")
        if not self.has_table("trackers"):
            self.action("CREATE TABLE trackers (provider TEXT, time NUMERIC, trackers TEXT)")

        self.provider = provider_instance
        self.provider_id = provider_instance.get_id()
        self.update_frequency = update_frequency

    def get_trackers(self):
        sql_result = self.select_one("SELECT * FROM trackers WHERE provider = ?", [self.provider_id])
        if sql_result:
            last_time = datetime.datetime.fromtimestamp(sql_result["time"])
            if last_time > datetime.datetime.now():
                last_time = datetime.datetime.min
        else:
            last_time = datetime.datetime.min

        if datetime.datetime.now() - last_time > self.update_frequency:
            trackers = self.provider.get_tracker_list()

            self.upsert(
                "trackers",
                {"time": int(time.mktime(datetime.datetime.now().timetuple())), "trackers": trackers},
                {"provider": self.provider_id},
            )
            result = trackers
        else:
            result = sql_result["trackers"]

        return result



import datetime
import itertools
import time
import traceback

from sickchill import logger, settings
from sickchill.helper.exceptions import AuthException
from sickchill.show.Show import Show

from . import db, show_name_helpers
from .databases import cache
from .name_parser.parser import InvalidNameException, InvalidShowException, NameParser
from .rssfeeds import getFeed

provider_cache_db = {}


class CacheDBConnection(db.DBConnection):
    def __init__(self):
        super().__init__("cache.db")
        db.upgrade_database(self, cache.InitialSchema)
        self.action("DELETE from results WHERE added < datetime('now','-30 days')")


class TVCache(object):
    def __init__(self, provider, **kwargs):
        self.provider = provider
        self.provider_id = self.provider.get_id()
        self.provider_db = None
        self.min_time = kwargs.pop("min_time", 10)
        self.search_params = kwargs.pop("search_params", dict(RSS=[""]))

    def _get_db(self):
        # init provider database if not done already
        if not provider_cache_db.get("instance"):
            provider_cache_db["instance"] = CacheDBConnection()

        return provider_cache_db.get("instance")

    def _clear_cache(self):
        if self.should_clear_cache():
            cache_db_con = self._get_db()
            cache_db_con.action("DELETE FROM results WHERE provider = ? AND added < datetime('now', '-30 days')", [self.provider_id])

    def _get_seeders_and_leechers(self, item):
        return self.provider._get_seeders_and_leechers(item)

    def _get_size(self, item):
        return self.provider._get_size(item)

    def _get_title_and_url(self, item):
        return self.provider._get_title_and_url(item)

    def _get_rss_data(self):
        return {"entries": self.provider.search(self.search_params)} if self.search_params else None

    def _check_auth(self, data):
        return True

    def update_cache(self):
        # check if we should update
        if not self.should_update():
            return

        try:
            data = self._get_rss_data()
            if self._check_auth(data):
                # clear cache
                self._clear_cache()

                # set updated
                self.set_last_update()

                cl = []
                for item in data["entries"] or []:
                    ci = self._parse_item(item)
                    if ci:
                        cl.append(ci)

                if cl:
                    cache_db_con = self._get_db()
                    cache_db_con.mass_upsert("results", cl)

        except AuthException as e:
            logger.warning("Authentication error: " + str(e))
        except Exception as e:
            logger.debug("Error while searching " + self.provider.name + ", skipping: " + repr(e))
            logger.debug(traceback.format_exc())

    def get_rss_feed(self, url, params=None):
        if self.provider.login():
            return getFeed(url, params=params, request_hook=self.provider.get_url)
        return {"entries": []}

    @staticmethod
    def _translate_title(title):
        return "" + title.replace(" ", ".")

    @staticmethod
    def _translate_link_url(url):
        return url.replace("&amp;", "&")

    def _parse_item(self, item):
        title, url = self._get_title_and_url(item)
        size = self._get_size(item)
        seeders, leechers = self._get_seeders_and_leechers(item)

        if title and url:
            title = self._translate_title(title)
            url = self._translate_link_url(url)

            # logger.debug("Attempting to add item to cache: " + title)
            return self._add_cache_entry(title, url, size, seeders, leechers)

        else:
            logger.debug("The data returned from the " + self.provider.name + " feed is incomplete, this result is unusable")

        return False

    @property
    def last_update(self):
        cache_db_con = self._get_db()
        sql_results = cache_db_con.select("SELECT time FROM lastUpdate WHERE provider = ?", [self.provider_id])

        if sql_results:
            last_time = int(sql_results[0]["time"])
            if last_time > int(time.mktime(datetime.datetime.today().timetuple())):
                last_time = 0
        else:
            last_time = 0

        return datetime.datetime.fromtimestamp(last_time)

    @property
    def last_search(self):
        cache_db_con = self._get_db()
        sql_results = cache_db_con.select("SELECT time FROM lastSearch WHERE provider = ?", [self.provider_id])

        if sql_results:
            last_time = int(sql_results[0]["time"])
            if last_time > int(time.mktime(datetime.datetime.today().timetuple())):
                last_time = 0
        else:
            last_time = 0

        return datetime.datetime.fromtimestamp(last_time)

    def set_last_update(self, to_date=None):
        """
        Sets the last update date for the current provider in the cache database

        :param to_date: date to set to, or None for today
        """
        if not to_date:
            to_date = datetime.datetime.today()

        cache_db_con = self._get_db()
        cache_db_con.upsert("lastUpdate", {"time": int(time.mktime(to_date.timetuple()))}, {"provider": self.provider_id})

    def set_last_search(self, to_date=None):
        """
        Sets the last search date for the current provider in the cache database

        :param to_date: date to set to, or None for today
        """
        if not to_date:
            to_date = datetime.datetime.today()

        cache_db_con = self._get_db()
        cache_db_con.upsert("lastSearch", {"time": int(time.mktime(to_date.timetuple()))}, {"provider": self.provider_id})

    def should_update(self):
        # if we've updated recently then skip the update
        if datetime.datetime.today() - self.last_update < datetime.timedelta(minutes=self.min_time):
            logger.debug("Last update was too soon, using old cache: " + str(self.last_update) + ". Updated less then " + str(self.min_time) + " minutes ago")
            return False

        return True

    def should_clear_cache(self):
        # if daily search hasn't used our previous results yet then don't clear the cache
        if self.last_update > self.last_search:
            return False

        return True

    def _add_cache_entry(self, name, url, size, seeders, leechers, parse_result=None, indexer_id=0):

        # check if we passed in a parsed result or should we try and create one
        if not parse_result:

            # create show_obj from indexer_id if available
            show_obj = None
            if indexer_id:
                show_obj = Show.find(settings.showList, indexer_id)

            try:
                parse_result = NameParser(showObj=show_obj).parse(name)
            except (InvalidNameException, InvalidShowException) as error:
                logger.debug("{0}".format(error))
                return None

            if not parse_result or not parse_result.series_name:
                return None

        # if we made it this far then lets add the parsed result to cache for usage later on
        season = parse_result.season_number if parse_result.season_number else 1
        episodes = parse_result.episode_numbers

        if season and episodes:
            # store episodes as a separated string
            episode_text = "|" + "|".join({str(episode) for episode in sorted(episodes) if episode}) + "|"

            # get the current timestamp
            cur_timestamp = int(time.mktime(datetime.datetime.today().timetuple()))

            # get quality of release
            quality = parse_result.quality

            # get release group
            release_group = parse_result.release_group

            # get version
            version = parse_result.version

            logger.debug(_("Added RSS item: [{}] to cache: {}").format(name, self.provider_id))
            return (
                {
                    "provider": self.provider_id,
                    "name": name,
                    "season": season,
                    "episodes": episode_text,
                    "indexerid": parse_result.show.indexerid,
                    "url": url,
                    "time": cur_timestamp,
                    "quality": quality,
                    "release_group": release_group,
                    "version": version,
                    "seeders": seeders,
                    "leechers": leechers,
                    "size": size,
                },
                {"url": url},
            )

    def search_cache(self, episode, manual_search=False, down_cur_quality=False):
        needed_eps = self.find_needed_episodes(episode, manual_search, down_cur_quality)
        return needed_eps.get(episode, [])

    def list_propers(self, date=None):
        cache_db_con = self._get_db()
        sql = "SELECT * FROM results WHERE provider = ? AND name LIKE '%.PROPER.%' OR name LIKE '%.REPACK.%'"
        # Add specific provider proper_strings also, like REAL, RERIP, etc.
        if hasattr(self.provider, "proper_strings"):
            if self.provider.proper_strings:
                for item in self.provider.proper_strings:
                    if "|" in item:
                        items = item.split("|")
                        for _item in items:
                            if _item.upper() not in sql:
                                sql += " OR name LIKE '%.{}.%'".format(_item)
                    elif item.upper() not in sql:
                        sql += " OR name LIKE '%.{}.%'".format(item)

        if date is not None:
            sql += " AND time >= " + str(int(time.mktime(date.timetuple())))

        propers_results = cache_db_con.select(sql, [self.provider_id])
        return [x for x in propers_results if x["indexerid"]]

    def find_needed_episodes(self, episode, manualSearch=False, downCurQuality=False):
        needed_eps = {}
        cl = []

        cache_db_con = self._get_db()
        if not episode:
            sql_results = cache_db_con.select("SELECT * FROM results WHERE provider = ?", [self.provider_id])
        elif not isinstance(episode, list):
            sql_results = cache_db_con.select(
                "SELECT * FROM results WHERE provider = ? AND indexerid = ? AND season = ? AND episodes LIKE ?",
                [self.provider_id, episode.show.indexerid, episode.season, "%|" + str(episode.episode) + "|%"],
            )
        else:
            for ep_obj in episode:
                cl.append(
                    [
                        "SELECT * FROM results WHERE provider = ? AND indexerid = ? AND season = ? AND episodes LIKE ? AND quality IN ("
                        + ",".join([str(x) for x in ep_obj.wantedQuality])
                        + ")",
                        [self.provider_id, ep_obj.show.indexerid, ep_obj.season, "%|" + str(ep_obj.episode) + "|%"],
                    ]
                )

            sql_results = cache_db_con.mass_action(cl, fetchall=True)
            sql_results = list(itertools.chain(*sql_results))

        # for each cache entry
        for cur_result in sql_results:
            # get the show object, or if it's not one of our shows then ignore it
            show_obj = Show.find(settings.showList, int(cur_result["indexerid"]))
            if not show_obj:
                continue

            # ignored/required words, and non-tv junk
            if not show_name_helpers.filter_bad_releases(cur_result["name"], show=show_obj):
                continue

            # skip if provider is anime only and show is not anime
            if self.provider.anime_only and not show_obj.is_anime:
                logger.debug("" + str(show_obj.name) + " is not an anime, skiping")
                continue

            # get season and ep data (ignoring multi-eps for now)
            cur_season = int(cur_result["season"])
            if cur_season == -1:
                continue

            cur_ep = cur_result["episodes"].split("|")[1]
            if not cur_ep:
                continue

            cur_ep = int(cur_ep)

            cur_quality = int(cur_result["quality"])
            cur_release_group = cur_result["release_group"]
            cur_version = cur_result["version"]

            # if the show says we want that episode then add it to the list
            if not show_obj.wantEpisode(cur_season, cur_ep, cur_quality, manualSearch, downCurQuality):
                logger.debug("Ignoring " + cur_result["name"])
                continue

            ep_obj = show_obj.getEpisode(cur_season, cur_ep)

            # build a result object
            title = cur_result["name"]
            url = cur_result["url"]

            logger.info("Found result " + title + " at " + url)

            result = self.provider.get_result([ep_obj])
            result.show = show_obj
            result.url = url
            result.name = title
            result.quality = cur_quality
            result.release_group = cur_release_group
            result.version = cur_version
            result.content = None

            # add it to the list
            if ep_obj not in needed_eps:
                needed_eps[ep_obj] = [result]
            else:
                needed_eps[ep_obj].append(result)

        # datetime stamp this search so cache gets cleared
        self.set_last_search()

        return needed_eps



from urllib.parse import urljoin

from sickchill import logger, settings
from sickchill.helper.common import try_int
from sickchill.oldbeard import tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("EZTV")

        # Credentials
        self.public = True

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0

        # URLs
        self.url = "https://eztv.re"
        self.api = urljoin(self.url, "api/get-torrents")

        # Cache
        self.cache = tvcache.TVCache(self, min_time=30)  # only poll ThePirateBay every 30 minutes max

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        search_params = {"imdb_id": None, "page": 1, "limit": 100}

        # Just doing the first page of results, because there is no search filter
        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            if mode != "RSS":
                if not (self.show and self.show.imdb_id):
                    continue

                search_params["imdb_id"] = self.show.imdb_id.strip("tt")
                logger.debug("Search string: {}".format(self.show.imdb_id))
            else:
                search_params.pop("imdb_id")

            data = self.get_url(self.api, params=search_params, returns="json")

            if not (data and isinstance(data, dict) and "torrents" in data):
                logger.debug("URL did not return data")
                continue

            for result in data["torrents"]:
                try:
                    title = result["title"][0:-5].replace(" ", ".")
                    info_hash = result["hash"]
                    if not all([title, info_hash]):
                        continue

                    link = result[("magnet_url", "torrent_url")[settings.TORRENT_METHOD == "blackhole"]]

                    seeders = result["seeds"]
                    leechers = result["peers"]

                    # Filter unseeded torrent
                    if seeders < self.minseed or leechers < self.minleech:
                        if mode != "RSS":
                            logger.debug(
                                "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers)
                            )
                        continue

                    torrent_size = try_int(result["size_bytes"])

                    item = {"title": title, "link": link, "size": torrent_size, "seeders": seeders, "leechers": leechers, "hash": info_hash}
                    if mode != "RSS":
                        logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                    items.append(item)
                except Exception:
                    continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



from urllib.parse import urljoin

import validators

from sickchill import logger
from sickchill.helper.common import try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("SkyTorrents")

        self.public = True

        self.minseed = 0
        self.minleech = 0

        self.url = "https://www.skytorrents.org"
        # https://www.skytorrents.org/?query=arrow&category=show&tag=hd&sort=seeders&type=video
        # https://www.skytorrents.org/top100?category=show&type=video&sort=created
        self.urls = {"search": urljoin(self.url, "/"), "rss": urljoin(self.url, "/top100")}

        self.custom_url = None

        self.cache = tvcache.TVCache(self, search_params={"RSS": [""]})

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                search_url = (self.urls["search"], self.urls["rss"])[mode == "RSS"]
                if self.custom_url:
                    if not validators.url(self.custom_url):
                        logger.warning("Invalid custom url: {0}".format(self.custom_url))
                        return results
                    search_url = urljoin(self.custom_url, search_url.split(self.url)[1])

                if mode != "RSS":
                    search_params = {"search": search_string, "sort": ("seeders", "created")[mode == "RSS"]}
                else:
                    search_params = {"category": "show", "type": "video", "sort": "created"}

                data = self.get_url(search_url, params=search_params, returns="text")
                if not data:
                    logger.debug("Data returned from provider does not contain any torrents")
                    continue

                with BS4Parser(data, "html5lib") as html:
                    labels = [label.get_text(strip=True) for label in html("th")]
                    for item in html("tr", attrs={"data-size": True}):
                        try:
                            size = try_int(item["data-size"])
                            cells = item.find_all("td")

                            title_block_links = cells[labels.index("Name")].find_all("a")
                            title = title_block_links[0].get_text(strip=True)
                            info_hash = title_block_links[0]["href"].split("/")[1]
                            download_url = title_block_links[2]["href"]

                            seeders = try_int(cells[labels.index("Seeders")].get_text(strip=True))
                            leechers = try_int(cells[labels.index("Leechers")].get_text(strip=True))

                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": info_hash}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)

                        except (AttributeError, TypeError, KeyError, ValueError):
                            continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results



import re

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("TvChaosUK")

        self.username = None
        self.password = None

        self.minseed = 0
        self.minleech = 0
        self.freeleech = None

        self.url = "https://www.tvchaosuk.com/"
        self.urls = {"login": self.url + "takelogin.php", "index": self.url + "index.php", "search": self.url + "browse.php"}

        self.cache = tvcache.TVCache(self)

    def _check_auth(self):
        if self.username and self.password:
            return True

        raise AuthException("Your authentication credentials for " + self.name + " are missing, check your config.")

    def login(self):
        if len(self.session.cookies) >= 4:
            return True

        login_params = {"username": self.username, "password": self.password, "logout": "no", "submit": "LOGIN", "returnto": "/browse.php"}

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("Error: Username or password incorrect!", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        # Search Params
        search_params = {"do": "search", "search_type": "t_name", "category": 0, "include_dead_torrents": "no", "submit": "search"}

        # Units
        units = ["B", "KB", "MB", "GB", "TB", "PB"]

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            for search_string in {*search_strings[mode]}:

                if mode == "Season":
                    search_string = re.sub(r"(.*)S0?", r"\1Series ", search_string)

                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                search_params["keywords"] = search_string
                data = self.get_url(self.urls["search"], post_data=search_params, returns="text")
                if not data:
                    logger.debug(_("No data returned from provider"))
                    continue

                with BS4Parser(data, "html5lib") as html:
                    torrent_table = html.find(id="sortabletable")
                    torrent_rows = torrent_table("tr") if torrent_table else []

                    # Continue only if at least one Release is found
                    if len(torrent_rows) < 2:
                        logger.debug("Data returned from provider does not contain any torrents")
                        continue

                    labels = [label.img["title"] if label.img else label.get_text(strip=True) for label in torrent_rows[0]("td")]
                    for torrent in torrent_rows[1:]:
                        try:
                            if self.freeleech and not torrent.find("img", alt=re.compile("Free Torrent")):
                                continue

                            title = torrent.find(class_="tooltip-content").div.get_text(strip=True)
                            download_url = torrent.find(title="Click to Download this Torrent!").parent["href"]
                            if not all([title, download_url]):
                                continue

                            seeders = try_int(torrent.find(title="Seeders").get_text(strip=True))
                            leechers = try_int(torrent.find(title="Leechers").get_text(strip=True))

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the"
                                        " minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers)
                                    )
                                continue

                            # Chop off tracker/channel prefix or we cant parse the result!
                            if mode != "RSS" and search_params["keywords"]:
                                show_name_first_word = re.search(r"^[^ .]+", search_params["keywords"]).group()
                                if not title.startswith(show_name_first_word):
                                    title = re.sub(r".*(" + show_name_first_word + ".*)", r"\1", title)

                            # Change title from Series to Season, or we can't parse
                            if mode == "Season":
                                title = re.sub(r"(.*)(?i)Series", r"\1Season", title)

                            # Strip year from the end or we can't parse it!
                            title = re.sub(r"(.*)[. ]?\(\d{4}\)", r"\1", title)
                            title = re.sub(r"\s+", r" ", title)

                            torrent_size = torrent("td")[labels.index("Size")].get_text(strip=True)
                            size = convert_size(torrent_size, units=units) or -1

                            if mode != "RSS":
                                logger.debug(
                                    _(
                                        "Found result: {title} with {seeders} seeders and {leechers} leechers".format(
                                            title=title, seeders=seeders, leechers=leechers
                                        )
                                    )
                                )

                            item = {"title": title + ".hdtv.x264", "link": download_url, "size": size, "seeders": seeders, "leechers": leechers}
                            items.append(item)
                        except Exception:
                            continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



import binascii

from enzyme import MKV
from pkg_resources import DistributionNotFound, get_distribution

import sickchill.oldbeard

try:
    get_distribution("pymediainfo")
    from pymediainfo import MediaInfo as mediainfo
except (ImportError, DistributionNotFound):
    mediainfo = None


def _avi_screen_size(filename):
    """
    Parses avi file header for width and height
    :param filename: full path and filename to a video file
    :type: str
    :returns tuple: (width, height)
    """
    try:
        if not filename.endswith(".avi"):
            with open(filename, "rb") as f:
                header = f.read(72)

            x = binascii.hexlify(header[68:72])
            height = int(x[6:8] + x[4:6] + x[2:4] + x[0:2], 16)
            assert 100 < height < 4320

            x = binascii.hexlify(header[64:68])
            width = int(x[6:8] + x[4:6] + x[2:4] + x[0:2], 16)
            assert 100 < width < 7680

            return width, height
    except Exception:
        pass

    return None, None


def _mkv_screen_size(filename):
    """
    Parses mkv file for width and height
    :param filename: full path and filename to a video file
    :type: str
    :returns tuple: (width, height)
    """
    try:
        if filename.endswith(".mkv"):
            with open(filename, "rb") as f:
                mkv = MKV(f)

            return mkv.video_tracks[0].width, mkv.video_tracks[0].height
    except Exception:
        pass

    return None, None


def _mediainfo_screen_size(filename):
    """
    Attempts to read the width and height of a video file, using mediainfo
    :param filename: full path and filename to a video file
    :type: str
    :returns tuple: (width, height)
    """
    try:
        if mediainfo:
            _media_info = mediainfo.parse(filename)
            for track in _media_info.tracks:
                if track.track_type == "Video":
                    return track.width, track.height
    except (OSError, TypeError):
        pass

    return None, None


# Only try to parse processable files once. Resets on restart ofc
bad_files = set()


def video_screen_size(filename):
    """
    Attempts to read the width and height of a video file,
    first using mediainfo and then enzyme, and then a custom avi reader

    :param filename: full path and filename to a video file
    :type: str
    :returns tuple: (width, height)
    """

    if filename in bad_files or not sickchill.oldbeard.helpers.is_media_file(filename):
        return None, None

    # Need to implement mediainfo another way, pymediainfo 2.0 causes segfaults
    # for method in [_mediainfo_screen_size, _mkv_screen_size, _avi_screen_size]:
    for method in [_mkv_screen_size, _avi_screen_size]:

        screen_size = method(filename)
        if screen_size != (None, None):
            return screen_size

    bad_files.add(filename)
    return None, None
